import json
from pathlib import Path
from urllib import request
from urllib.error import URLError

from packaging import version

from LabGym import __version__, gui


def main() -> None:
    """Check version and run LabGym."""

    try:
        current_version = version.parse(__version__)
        pypi_json = json.loads(
            request.urlopen("https://pypi.python.org/pypi/LabGym/json").read()
        )
        latest_version = version.parse(pypi_json["info"]["version"])

        if latest_version > current_version:
            if "pipx" in str(Path(__file__)):
                upgrade_command = "pipx upgrade LabGym"
            else:
                upgrade_command = "python3 -m pip install --upgrade LabGym"

            print(
                f"You are using LabGym {current_version}, but version {latest_version} is available."
            )
            print(
                f"Consider upgrading LabGym by using the command '{upgrade_command}'."
            )
            print(
                "For the details of new changes, check https://github.com/umyelab/LabGym.\n"
            )
    except URLError:
        print("Unable to check for new versions of LabGym!")
        print("Please check https://github.com/umyelab/LabGym for updates.\n")

    gui.gui()


if __name__ == "__main__":
    main()



"""
Copyright (C)
This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program. If not, see https://tldrlegal.com/license/gnu-general-public-license-v3-(gpl-3)#fulltext. 

For license issues, please contact:

Dr. Bing Ye
Life Sciences Institute
University of Michigan
210 Washtenaw Avenue, Room 5403
Ann Arbor, MI 48109-2216
USA

Email: bingye@umich.edu
"""


from ..utils import LabGymWindow
from .analyze_behaviors import AnalyzeBehaviors
from .mine_results import MineResults


class AnalysisModule(LabGymWindow):
    def __init__(self):
        super().__init__(title="Analysis Module", size=(500, 220))

        self.add_submit_button(
            label="Analyze Behaviors",
            handler=self.analyze_behaviors,
            tool_tip="Automatically track animals / objects of your interest, identify and quantify their behaviors in videos.",
        )

        self.add_submit_button(
            label="Mine Results",
            handler=self.mine_results,
            tool_tip="Automatically mine the analysis results to display the data details that show statistically significant differences among groups of your selection.",
        )

        self.display_window()

    def analyze_behaviors(self, event):
        AnalyzeBehaviors()

    def mine_results(self, event):
        MineResults()



"""
Copyright (C)
This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program. If not, see https://tldrlegal.com/license/gnu-general-public-license-v3-(gpl-3)#fulltext.

For license issues, please contact:

Dr. Bing Ye
Life Sciences Institute
University of Michigan
210 Washtenaw Avenue, Room 5403
Ann Arbor, MI 48109-2216
USA

Email: bingye@umich.edu
"""

from __future__ import annotations

import os
from collections import deque
from pathlib import Path
from typing import Tuple

import cv2
import numpy as np
import wx
from cv2.typing import MatLike

from LabGym.analyzebehaviors import AnalyzeAnimal
from LabGym.analyzebehaviorsdetector import (
    AnalyzeAnimalDetector,
    get_animal_names,
    get_detector_names,
)
from LabGym.gui.utils import (
    WX_IMAGE_WILDCARD,
    WX_VIDEO_WILDCARD,
    BehaviorMode,
    LabGymWindow,
)
from LabGym.tools import AnimalVsBg

THE_ABSOLUTE_CURRENT_PATH = str(Path(__file__).resolve().parent.parent.parent)


class GenerateBehaviorExamples(LabGymWindow):
    """Generate behavior examples for the user to sort."""

    def __init__(self):
        super().__init__(title="Generate Behavior Examples", size=(1000, 530))
        self.behavior_mode = BehaviorMode.NON_INTERACTIVE
        self.use_detector = False
        self.detector_path = None
        self.path_to_detector = None
        self.detection_threshold = 0
        self.animal_kinds = []
        self.background_path = None
        self.path_to_videos = None
        self.result_path = None
        self.framewidth = None
        self.delta = 10000
        self.decode_animalnumber = False
        self.animal_number = None
        self.autofind_t = False
        self.decode_t = False
        self.t = 0
        self.duration = 0
        self.decode_extraction = False
        self.ex_start = 0
        self.ex_end = None
        self.animal_vs_bg = AnimalVsBg.ANIMAL_LIGHTER
        self.stable_illumination = True
        self.length = 15
        self.skip_redundant = 1
        self.include_bodyparts = False
        self.std = 0
        self.background_free = True
        self.social_distance = 0

        self.text_specifymode = self.module_text(
            "Default: Non-interactive: behaviors of each individuals (each example contains one animal / object)"
        )
        self.add_module(
            "Specify the mode of behavior\nexamples to generate",
            self.specify_mode,
            '"Non-interactive" is for behaviors of each individual; "Interactive basic" is for interactive behaviors of all animals but not distinguishing each individual; "Interactive advanced" is slower in analysis than "basic" but distinguishes individuals during close body contact. "Static images" is for analyzing images not videos. See Extended Guide for details.',
            self.text_specifymode,
        )

        self.text_inputvideos = self.module_text("None.")
        self.add_module(
            "Select the video(s) / image(s) to\ngenerate behavior examples",
            self.select_videos,
            "Select one or more videos / images. Common video formats (mp4, mov, avi, m4v, mkv, mpg, mpeg) or image formats (jpg, jpeg, png, tiff, bmp) are supported except wmv format.",
            self.text_inputvideos,
        )

        self.text_outputfolder = self.module_text("None.")
        self.add_module(
            "Select a folder to store the\ngenerated behavior examples",
            self.select_outpath,
            'Will create a subfolder for each video in the selected folder. Each subfolder is named after the file name of the video and stores the generated behavior examples. For "Static images" mode, all generated behavior examples will be in this folder.',
            self.text_outputfolder,
        )

        self.text_detection = self.module_text(
            "Default: Background subtraction-based method."
        )
        self.add_module(
            "Specify the method to\ndetect animals or objects",
            self.select_method,
            "Background subtraction-based method is accurate and fast but needs static background and stable illumination in videos; Detectors-based method is accurate and versatile in any recording settings but is slow. See Extended Guide for details.",
            self.text_detection,
        )

        self.text_startgenerate = self.module_text(
            "Default: at the beginning of the video(s)."
        )
        self.add_module(
            "Specify when generating behavior examples\nshould begin (unit: second)",
            self.specify_timing,
            'Enter a beginning time point for all videos or use "Decode from filenames" to let LabGym decode the different beginning time for different videos. See Extended Guide for details.',
            self.text_startgenerate,
        )

        self.text_duration = self.module_text(
            "Default: from the specified beginning time to the end of a video."
        )
        self.add_module(
            "Specify how long generating examples\nshould last (unit: second)",
            self.input_duration,
            "The duration is the same for all the videos in one batch.",
            self.text_duration,
        )

        self.text_animalnumber = self.module_text("Default: 1.")
        self.add_module(
            "Specify the number of animals\nin a video",
            self.specify_animalnumber,
            'Enter a number for all videos or use "Decode from filenames" to let LabGym decode the different animal number for different videos. See Extended Guide for details.',
            self.text_animalnumber,
        )

        self.text_length = self.module_text("Default: 15 frames.")
        self.add_module(
            "Specify the number of frames for\nan animation / pattern image",
            self.input_length,
            "The duration (the number of frames, an integer) of each behavior example, which should approximate the length of a behavior episode. This duration needs to be the same across all the behavior examples for training one Categorizer. See Extended Guide for details.",
            self.text_length,
        )

        self.text_skipredundant = self.module_text(
            "Default: no frame to skip (generate a behavior example every frame)."
        )
        self.add_module(
            "Specify how many frames to skip when\ngenerating two consecutive behavior examples",
            self.specify_redundant,
            "If two consecutively generated examples have many overlapping frames, they look similar, which makes training inefficient and sorting laborious. Specifying an interval (skipped frames) between two examples can address this. See Extended Guide for details.",
            self.text_skipredundant,
        )

        self.add_submit_button(
            "Start to generate behavior examples",
            self.generate_data,
            "Need to specify whether to include background and body parts in the generated behavior examples. See Extended Guide for details.",
        )

        self.display_window()

    def specify_mode(self, event):
        """Select a behavior mode to generate examples."""

        # The indices of each behavior mode in this list are mapped to the same
        # constant values used in the BehaviorMode class.
        behavior_modes = [
            "Non-interactive: behaviors of each individual (each example contains one animal / object)",
            "Interactive basic: behaviors of all (each example contains all animals / objects)",
            "Interactive advanced: behaviors of each individual and social groups (each example contains either one or multiple animals / objects)",
            "Static images (non-interactive): behaviors of each individual in static images(each image contains one animal / object)",
        ]
        dialog = wx.SingleChoiceDialog(
            self,
            message="Specify the mode of behavior examples to generate",
            caption="Behavior-example mode",
            choices=behavior_modes,
        )
        if dialog.ShowModal() != wx.ID_OK:
            self.behavior_mode = BehaviorMode.NON_INTERACTIVE
            self.text_specifymode.SetLabel(
                f"Behavior mode: {behavior_modes[self.behavior_mode]}."
            )
            dialog.Destroy()
            return

        self.behavior_mode = dialog.GetSelection()
        dialog.Destroy()

        if self.behavior_mode == BehaviorMode.INTERACT_ADVANCED:
            # Need to set social distance for Interactive Advanced mode
            dialog1 = wx.NumberEntryDialog(
                self,
                "Interactions happen within the social distance.",
                "(See Extended Guide for details)\nHow many folds of square root of the animals area\nis the social distance (0=infinity far):",
                "Social distance (Enter an integer)",
                0,
                0,
                100000000000000,
            )
            if dialog1.ShowModal() == wx.ID_OK:
                self.social_distance = int(dialog1.GetValue())
                if self.social_distance == 0:
                    self.social_distance = float("inf")
            else:
                self.social_distance = float("inf")
            dialog1.Destroy()
            self.text_detection.SetLabel(
                "Only Detector-based detection method is available for the selected behavior mode."
            )
        elif self.behavior_mode == BehaviorMode.STATIC_IMAGES:
            self.text_detection.SetLabel(
                "Only Detector-based detection method is available for the selected behavior mode."
            )
            self.text_startgenerate.SetLabel(
                'No need to specify this since the selected behavior mode is "Static images".'
            )
            self.text_duration.SetLabel(
                'No need to specify this since the selected behavior mode is "Static images".'
            )
            self.text_animalnumber.SetLabel(
                'No need to specify this since the selected behavior mode is "Static images".'
            )
            self.text_length.SetLabel(
                'No need to specify this since the selected behavior mode is "Static images".'
            )
            self.text_skipredundant.SetLabel(
                'No need to specify this since the selected behavior mode is "Static images".'
            )

        label_text = f"Behavior mode: {behavior_modes[self.behavior_mode]}"
        if self.behavior_mode == BehaviorMode.INTERACT_ADVANCED:
            label_text += f" with social distance {self.social_distance}"
        label_text += "."
        self.text_specifymode.SetLabel(label_text)

    def select_videos(self, event):
        """Select videos/images to generate behavior examples from."""
        if self.behavior_mode == BehaviorMode.STATIC_IMAGES:
            wildcard = WX_IMAGE_WILDCARD
        else:
            wildcard = WX_VIDEO_WILDCARD

        dialog = wx.FileDialog(
            self, "Select video(s) / image(s)", "", "", wildcard, style=wx.FD_MULTIPLE
        )
        if dialog.ShowModal() != wx.ID_OK:
            dialog.Destroy()
            return

        self.path_to_videos = dialog.GetPaths()
        self.path_to_videos.sort()
        dialog.Destroy()

        folder = os.path.dirname(self.path_to_videos[0])
        dialog1 = wx.MessageDialog(
            self,
            'Proportional resize the video frames / images?\nSelect "No" if dont know what it is.',
            "(Optional) resize the frames / images?",
            wx.YES_NO | wx.ICON_QUESTION,
        )
        if dialog1.ShowModal() != wx.ID_YES:
            self.framewidth = None
            self.text_inputvideos.SetLabel(
                f"Selected {len(self.path_to_videos)} file(s) in {folder} (original frame / image size)."
            )
            dialog1.Destroy()
            return

        dialog2 = wx.NumberEntryDialog(
            self,
            "Enter the desired frame / image width",
            "The unit is pixel:",
            "Desired frame / image width",
            480,
            1,
            10000,
        )
        if dialog2.ShowModal() != wx.ID_OK:
            self.framewidth = None
            self.text_inputvideos.SetLabel(
                f"Selected {len(self.path_to_videos)} file(s) in {folder} (original frame / image size)."
            )
            dialog2.Destroy()
            return

        self.framewidth = max(int(dialog2.GetValue()), 10)
        dialog2.Destroy()
        self.text_inputvideos.SetLabel(
            f"Selected {len(self.path_to_videos)} file(s) in {folder} (proportionally resize frame / image width to {self.framewidth})."
        )

    def select_outpath(self, event):
        """Select folder to store behavior examples."""
        dialog = wx.DirDialog(self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE)
        if dialog.ShowModal() == wx.ID_OK:
            self.result_path = dialog.GetPath()
            self.text_outputfolder.SetLabel(
                f"Generate behavior examples in: {self.result_path}."
            )
        dialog.Destroy()

    def _configure_background_subtraction(self):
        """Configure background subtraction-based example generation."""
        self.use_detector = False

        # The indices of these options are mapped to the same values as the
        # constants in the AnimalVsBg class
        contrasts = [
            "Animal brighter than background",
            "Animal darker than background",
            "Hard to tell",
        ]
        dialog = wx.SingleChoiceDialog(
            self,
            message="Select the scenario that fits your videos best",
            caption="Which fits best?",
            choices=contrasts,
        )
        if dialog.ShowModal() != wx.ID_OK:
            dialog.Destroy()
            return

        self.animal_vs_bg = dialog.GetSelection()
        label_options = [
            "animal brighter",
            "animal darker",
            "animal partially brighter/darker",
        ]
        label_text = f"Background Subtraction: {label_options[self.animal_vs_bg]}"
        dialog.Destroy()

        dialog2 = wx.MessageDialog(
            self,
            'Load an existing background from a folder?\nSelect "No" if dont know what it is.',
            "(Optional) load existing background?",
            wx.YES_NO | wx.ICON_QUESTION,
        )
        if dialog2.ShowModal() == wx.ID_YES:
            start_time_dialog = wx.DirDialog(
                self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE
            )
            if start_time_dialog.ShowModal() == wx.ID_OK:
                self.background_path = start_time_dialog.GetPath()
            start_time_dialog.Destroy()
        else:
            self.background_path = None
            if self.animal_vs_bg != AnimalVsBg.HARD_TO_TELL:
                start_time_dialog = wx.MessageDialog(
                    self,
                    'Unstable illumination in the video?\nSelect "Yes" if dont know what it is.',
                    "(Optional) unstable illumination?",
                    wx.YES_NO | wx.ICON_QUESTION,
                )
                self.stable_illumination = start_time_dialog.ShowModal() != wx.ID_YES
                start_time_dialog.Destroy()
        dialog2.Destroy()

        if self.background_path is not None:
            self.text_detection.SetLabel(label_text)
            return

        extraction_methods = [
            "Use the entire duration (default but NOT recommended)",
            'Decode from filenames: "_xst_" and "_xet_"',
            "Enter two time points",
        ]
        dialog2 = wx.SingleChoiceDialog(
            self,
            message="Specify the time window for background extraction",
            caption="Time window for background extraction",
            choices=extraction_methods,
        )
        if dialog2.ShowModal() != wx.ID_OK:
            self.text_detection.SetLabel(label_text)
            return

        extraction_method = dialog2.GetStringSelection()
        dialog2.Destroy()

        self.decode_extraction = False
        if extraction_method == extraction_methods[0]:
            self.text_detection.SetLabel(f"{label_text}, using the entire duration.")
            return
        elif extraction_method == extraction_methods[1]:
            self.decode_extraction = True
            self.text_detection.SetLabel(
                f"{label_text}, using time window decoded from filenames '_xst_' and '_xet_'."
            )
            return

        # Prompt for extraction time window
        label_text += ", using time window (in seconds) from "
        start_time_dialog = wx.NumberEntryDialog(
            self,
            "Enter the start time",
            "The unit is second:",
            "Start time for background extraction",
            0,
            0,
            100000000000000,
        )
        if start_time_dialog.ShowModal() == wx.ID_OK:
            self.ex_start = int(start_time_dialog.GetValue())
            label_text += f"{self.ex_start} to "
        start_time_dialog.Destroy()

        end_time_dialog = wx.NumberEntryDialog(
            self,
            "Enter the end time",
            "The unit is second:",
            "End time for background extraction",
            0,
            0,
            100000000000000,
        )
        if end_time_dialog.ShowModal() == wx.ID_OK:
            self.ex_end = int(end_time_dialog.GetValue())
            if self.ex_end == 0:
                self.ex_end = None
                label_text += "the end."
            else:
                label_text += f"{self.ex_end}."
        end_time_dialog.Destroy()

        self.text_detection.SetLabel(label_text)

    def _configure_detector(self):
        """Configure Detector-based example generation."""
        self.use_detector = True
        self.animal_number = {}
        self.detector_path = os.path.join(THE_ABSOLUTE_CURRENT_PATH, "detectors")

        detectors = get_detector_names()
        detectors.append("Choose a new directory of the Detector")

        select_detector_dialog = wx.SingleChoiceDialog(
            self,
            message="Select a Detector for animal detection",
            caption="Select a Detector",
            choices=detectors,
        )
        if select_detector_dialog.ShowModal() != wx.ID_OK:
            select_detector_dialog.Destroy()
            return
        else:
            detector = select_detector_dialog.GetStringSelection()
            select_detector_dialog.Destroy()

        if detector == "Choose a new directory of the Detector":
            dialog2 = wx.DirDialog(
                self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE
            )
            if dialog2.ShowModal() == wx.ID_OK:
                self.path_to_detector = dialog2.GetPaths()
            dialog2.Destroy()
        else:
            self.path_to_detector = os.path.join(self.detector_path, detector)

        animal_names = get_animal_names(detector)
        if len(animal_names) > 1:
            dialog2 = wx.MultiChoiceDialog(
                self,
                message="Specify which animals/objects involved in behavior examples",
                caption="Animal/Object kind",
                choices=animal_names,
            )
            if dialog2.ShowModal() == wx.ID_OK:
                self.animal_kinds = [animal_names[i] for i in dialog2.GetSelections()]
            else:
                self.animal_kinds = animal_names
            dialog2.Destroy()
        else:
            self.animal_kinds = animal_names

        if self.behavior_mode == BehaviorMode.STATIC_IMAGES:
            dialog2 = wx.NumberEntryDialog(
                self,
                "Enter the Detector's detection threshold (0~100%)",
                "The higher detection threshold,\nthe higher detection accuracy,\nbut the lower detection sensitivity.\nEnter 0 if don't know how to set.",
                "Detection threshold",
                0,
                0,
                100,
            )
            if dialog2.ShowModal() == wx.ID_OK:
                detection_threshold = dialog2.GetValue()
                self.detection_threshold = detection_threshold / 100
                self.text_detection.SetLabel(
                    f"Detector: {detector} (detection threshold: {detection_threshold}%); The animals/objects: {self.animal_kinds}."
                )
            dialog2.Destroy()
        else:
            for animal_name in self.animal_kinds:
                self.animal_number[animal_name] = 1
            self.text_animalnumber.SetLabel(
                f"The number of {self.animal_kinds} is: {list(self.animal_number.values())}."
            )
            self.text_detection.SetLabel(
                f"Detector: {detector}; The animals/objects: {self.animal_kinds}."
            )

    def select_method(self, event):
        """Select method to generate contours for behavior examples."""
        DETECTOR = "Use trained Detectors (versatile but slow)"
        BACKGROUND_SUBTRACTION = "Subtract background (fast but requires static background & stable illumination)"
        methods = [DETECTOR]
        if self.behavior_mode in [
            BehaviorMode.NON_INTERACTIVE,
            BehaviorMode.INTERACT_BASIC,
        ]:
            methods.append(BACKGROUND_SUBTRACTION)

        dialog = wx.SingleChoiceDialog(
            self,
            message="How to detect the animals?",
            caption="Detection methods",
            choices=methods,
        )
        if dialog.ShowModal() != wx.ID_OK:
            dialog.Destroy()
            return

        method = dialog.GetStringSelection()
        dialog.Destroy()

        if method == BACKGROUND_SUBTRACTION:
            self._configure_background_subtraction()
        else:
            self._configure_detector()

    def specify_timing(self, event):
        """Specify how to handle lighting changes for optogenetic experiments."""
        if self.behavior_mode == BehaviorMode.STATIC_IMAGES:
            wx.MessageBox(
                'No need to specify this since the selected behavior mode is "Static images".',
                "Error",
                wx.OK | wx.ICON_ERROR,
            )
            return

        if self.use_detector is False:
            dialog = wx.MessageDialog(
                self,
                "light on and off in videos?",
                "Illumination shifts?",
                wx.YES_NO | wx.ICON_QUESTION,
            )
            if dialog.ShowModal() == wx.ID_YES:
                self.delta = 1.2
            else:
                self.delta = 10000
            dialog.Destroy()

        AUTOMATIC = "Automatic (for light on and off)"
        DECODE = 'Decode from filenames: "_bt_"'
        TIME_POINT = "Enter a time point"
        methods = [DECODE, TIME_POINT]
        if self.delta == 1.2 and self.use_detector is False:
            methods.insert(0, AUTOMATIC)

        dialog = wx.SingleChoiceDialog(
            self,
            message="Specify beginning time to generate behavior examples",
            caption="Beginning time for generator",
            choices=methods,
        )
        if dialog.ShowModal() != wx.ID_OK:
            dialog.Destroy()
            return
        else:
            method = dialog.GetStringSelection()
            dialog.Destroy()

        self.autofind_t = False
        self.decode_t = False
        if method == AUTOMATIC:
            self.autofind_t = True
            self.text_startgenerate.SetLabel(
                "Automatically find the onset of the 1st time when light on / off as the beginning time."
            )
        elif method == DECODE:
            self.decode_t = True
            self.text_startgenerate.SetLabel(
                'Decode from the filenames: the "t" immediately after the letter "b"" in "_bt_".'
            )
        else:
            dialog2 = wx.NumberEntryDialog(
                self,
                "Enter beginning time to generate examples",
                "The unit is second:",
                "Beginning time to generate examples",
                0,
                0,
                100000000000000,
            )
            if dialog2.ShowModal() == wx.ID_OK:
                self.t = max(float(dialog2.GetValue()), 0)
                self.text_startgenerate.SetLabel(
                    f"Start to generate behavior examples at the: {self.t} second."
                )
            dialog2.Destroy()

    def input_duration(self, event):
        """Select duration for generating behavior examples."""
        if self.behavior_mode == BehaviorMode.STATIC_IMAGES:
            wx.MessageBox(
                "No need to specify this since the selected behavior mode is 'Static images'.",
                "Error",
                wx.OK | wx.ICON_ERROR,
            )
            return

        dialog = wx.NumberEntryDialog(
            self,
            "Enter the duration for generating examples",
            "The unit is second:",
            "Duration for generating examples",
            0,
            0,
            100000000000000,
        )
        if dialog.ShowModal() == wx.ID_OK:
            self.duration = int(dialog.GetValue())
            if self.duration != 0:
                self.text_duration.SetLabel(
                    f"The generation of behavior examples lasts for {self.duration} seconds."
                )
            else:
                self.text_duration.SetLabel(
                    "The generation of behavior examples lasts for the entire duration of a video."
                )
        dialog.Destroy()

    def specify_animalnumber(self, event):
        """Select the number of animals in the video."""
        if self.behavior_mode == BehaviorMode.STATIC_IMAGES:
            wx.MessageBox(
                'No need to specify this since the selected behavior mode is "Static images".',
                "Error",
                wx.OK | wx.ICON_ERROR,
            )
            return

        DECODE = 'Decode from filenames: "_nn_"'
        ENTER = "Enter the number of animals"
        methods = [DECODE, ENTER]

        dialog = wx.SingleChoiceDialog(
            self,
            message="Specify the number of animals in a video",
            caption="The number of animals in a video",
            choices=methods,
        )
        if dialog.ShowModal() != wx.ID_OK:
            dialog.Destroy()
            return
        else:
            method = dialog.GetStringSelection()
            dialog.Destroy()

        self.decode_animalnumber = False
        if method == DECODE:
            self.decode_animalnumber = True
            self.text_animalnumber.SetLabel(
                'Decode from the filenames: the "n" immediately after the letter "n" in _"nn"_.'
            )
            return

        if self.use_detector:
            self.animal_number = {}
            for animal_name in self.animal_kinds:
                dialog1 = wx.NumberEntryDialog(
                    self,
                    "",
                    "The number of " + str(animal_name) + ": ",
                    str(animal_name) + " number",
                    1,
                    1,
                    100,
                )
                if dialog1.ShowModal() == wx.ID_OK:
                    self.animal_number[animal_name] = int(dialog1.GetValue())
                else:
                    self.animal_number[animal_name] = 1
                dialog1.Destroy()
            self.text_animalnumber.SetLabel(
                f"The number of {self.animal_kinds} is: {list(self.animal_number.values())}."
            )
        else:
            dialog1 = wx.NumberEntryDialog(
                self,
                "",
                "The number of animals:",
                "Animal number",
                1,
                1,
                100,
            )
            if dialog1.ShowModal() == wx.ID_OK:
                self.animal_number = int(dialog1.GetValue())
            else:
                self.animal_number = 1
            self.text_animalnumber.SetLabel(
                f"The total number of animals in a video is {self.animal_number}."
            )
            dialog1.Destroy()

    def input_length(self, event):
        """Enter the number of frames corresponding to a behavior."""
        if self.behavior_mode == BehaviorMode.STATIC_IMAGES:
            wx.MessageBox(
                'No need to specify this since the selected behavior mode is "Static images".',
                "Error",
                wx.OK | wx.ICON_ERROR,
            )
            return

        dialog = wx.NumberEntryDialog(
            self,
            "Enter the number of frames\nfor a behavior example",
            "Enter a number\n(minimum=3):",
            "Behavior episode duration",
            15,
            1,
            1000,
        )
        if dialog.ShowModal() == wx.ID_OK:
            self.length = max(int(dialog.GetValue()), 3)
            self.text_length.SetLabel(
                f"The duration of a behavior example is: {self.length} frames."
            )
        dialog.Destroy()

    def specify_redundant(self, event):
        """Select the number of frames to skip when generating examples."""
        if self.behavior_mode == BehaviorMode.STATIC_IMAGES:
            wx.MessageBox(
                'No need to specify this since the selected behavior mode is "Static images".',
                "Error",
                wx.OK | wx.ICON_ERROR,
            )
            return

        dialog = wx.NumberEntryDialog(
            self,
            "How many frames to skip?",
            "Enter a number:",
            "Interval for generating examples",
            15,
            0,
            100000000000000,
        )
        if dialog.ShowModal() == wx.ID_OK:
            self.skip_redundant = int(dialog.GetValue())
            self.text_skipredundant.SetLabel(
                f"Generate a pair of example every {self.skip_redundant} frames."
            )
        else:
            self.skip_redundant = 1
            self.text_skipredundant.SetLabel(
                "Generate a pair of example at every frame."
            )
        dialog.Destroy()

    def generate_data(self, event):
        """Generate behavior examples with the given configuration."""
        if self.path_to_videos is None or self.result_path is None:
            wx.MessageBox(
                "No input video(s) / output folder selected.",
                "Error",
                wx.OK | wx.ICON_ERROR,
            )
            return

        dialog = wx.MessageDialog(
            self,
            'Include background in animations? Select "No"\nif background is behavior irrelevant.',
            "Including background?",
            wx.YES_NO | wx.ICON_QUESTION,
        )
        self.background_free = dialog.ShowModal() != wx.ID_YES
        dialog.Destroy()

        if self.behavior_mode == BehaviorMode.STATIC_IMAGES:
            if self.path_to_detector is None:
                wx.MessageBox(
                    "You need to select a Detector.", "Error", wx.OK | wx.ICON_ERROR
                )
                return
            else:
                AAD = AnalyzeAnimalDetector()
                AAD.analyze_images_individuals(
                    self.path_to_detector,
                    self.path_to_videos,
                    self.result_path,
                    self.animal_kinds,
                    generate=True,
                    imagewidth=self.framewidth,
                    detection_threshold=self.detection_threshold,
                    background_free=self.background_free,
                )
                return

        dialog = wx.MessageDialog(
            self,
            'Include body parts in pattern images?\nSelect "No" if limb movement is neglectable.',
            "Including body parts?",
            wx.YES_NO | wx.ICON_QUESTION,
        )
        self.include_bodyparts = False
        if dialog.ShowModal() == wx.ID_YES:
            self.include_bodyparts = True
            dialog2 = wx.NumberEntryDialog(
                self,
                "Leave it as it is if dont know what it is.",
                "Enter a number between 0 and 255:",
                "STD for motionless pixels",
                0,
                0,
                255,
            )
            if dialog2.ShowModal() == wx.ID_OK:
                self.std = int(dialog2.GetValue())
            else:
                self.std = 0
            dialog2.Destroy()
        dialog.Destroy()

        dialog = wx.MessageDialog(
            self,
            "Start to generate behavior examples?",
            "Start to generate examples?",
            wx.YES_NO | wx.ICON_QUESTION,
        )
        if dialog.ShowModal() != wx.ID_YES:
            dialog.Destroy()
            return
        else:
            dialog.Destroy()

        for video in self.path_to_videos:
            filename = os.path.splitext(os.path.basename(video))[0].split("_")
            if self.decode_animalnumber is True:
                if self.use_detector is True:
                    self.animal_number = {}
                    number = [x[1:] for x in filename if len(x) > 1 and x[0] == "n"]
                    for a, animal_name in enumerate(self.animal_kinds):
                        self.animal_number[animal_name] = int(number[a])
                else:
                    for x in filename:
                        if len(x) > 1:
                            if x[0] == "n":
                                self.animal_number = int(x[1:])
            if self.decode_t is True:
                for x in filename:
                    if len(x) > 1:
                        if x[0] == "b":
                            self.t = float(x[1:])
            if self.decode_extraction is True:
                for x in filename:
                    if len(x) > 2:
                        if x[:2] == "xs":
                            self.ex_start = int(x[2:])
                        if x[:2] == "xe":
                            self.ex_end = int(x[2:])

            if self.animal_number is None:
                if self.use_detector is True:
                    self.animal_number = {}
                    for animal_name in self.animal_kinds:
                        self.animal_number[animal_name] = 1
                else:
                    self.animal_number = 1

            if self.use_detector is False:
                AA = AnalyzeAnimal()
                AA.prepare_analysis(
                    video,
                    self.result_path,
                    self.animal_number,
                    delta=self.delta,
                    framewidth=self.framewidth,
                    stable_illumination=self.stable_illumination,
                    channel=3,
                    include_bodyparts=self.include_bodyparts,
                    std=self.std,
                    categorize_behavior=False,
                    animation_analyzer=False,
                    path_background=self.background_path,
                    autofind_t=self.autofind_t,
                    t=self.t,
                    duration=self.duration,
                    ex_start=self.ex_start,
                    ex_end=self.ex_end,
                    length=self.length,
                    animal_vs_bg=self.animal_vs_bg,
                )
                if self.behavior_mode == BehaviorMode.NON_INTERACTIVE:
                    AA.generate_data(
                        background_free=self.background_free,
                        skip_redundant=self.skip_redundant,
                    )
                else:
                    AA.generate_data_interact_basic(
                        background_free=self.background_free,
                        skip_redundant=self.skip_redundant,
                    )
            else:
                AAD = AnalyzeAnimalDetector()
                AAD.prepare_analysis(
                    self.path_to_detector,
                    video,
                    self.result_path,
                    self.animal_number,
                    self.animal_kinds,
                    self.behavior_mode,
                    framewidth=self.framewidth,
                    channel=3,
                    include_bodyparts=self.include_bodyparts,
                    std=self.std,
                    categorize_behavior=False,
                    animation_analyzer=False,
                    t=self.t,
                    duration=self.duration,
                    length=self.length,
                    social_distance=self.social_distance,
                )
                if self.behavior_mode == BehaviorMode.NON_INTERACTIVE:
                    AAD.generate_data(
                        background_free=self.background_free,
                        skip_redundant=self.skip_redundant,
                    )
                elif self.behavior_mode == BehaviorMode.INTERACT_BASIC:
                    AAD.generate_data_interact_basic(
                        background_free=self.background_free,
                        skip_redundant=self.skip_redundant,
                    )
                else:
                    AAD.generate_data_interact_advance(
                        background_free=self.background_free,
                        skip_redundant=self.skip_redundant,
                    )


class SortBehaviorExamples(LabGymWindow):
    """Sort pattern images and animations."""

    def __init__(self):
        super().__init__(title="Sort Behavior Examples", size=(1000, 240))
        self.input_path = None
        self.result_path = None
        self.keys_behaviors = {}

        self.text_inputfolder = self.module_text("None.")
        self.add_module(
            "Select the folder that stores\nunsorted behavior examples",
            self.input_folder,
            'Select a folder that stores the behavior examples generated by "Generate Behavior Examples" functional unit. All examples in this folder should be in pairs (animation + pattern image).',
            self.text_inputfolder,
        )

        self.text_outputfolder = self.module_text("None.")
        self.add_module(
            "Select the folder to store\nthe sorted behavior examples",
            self.output_folder,
            "A subfolder will be created for each behavior type under the behavior name.",
            self.text_outputfolder,
        )

        self.text_keynames = self.module_text("None.")
        self.add_module(
            "Enter a single character shortcut key and\nthe corresponding behavior name",
            self.input_keys,
            'Format: "shortcutkey-behaviorname". "o", "p", "q", and "u" are reserved for "Previous", "Next", "Quit", and "Undo". When hit a shortcut key, the behavior example pair will be moved to the folder named after the corresponding behavior name.',
            self.text_keynames,
        )

        self.add_submit_button(
            "Sort behavior examples",
            self.sort_behaviors,
            "You will see each example pair in the screen one by one and can use shortcut keys to sort them into folders of the behavior types.",
        )

        self.display_window()

    def input_folder(self, event):
        """Select folder with unsorted behavior examples."""
        dialog = wx.DirDialog(self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE)
        if dialog.ShowModal() == wx.ID_OK:
            self.input_path = Path(dialog.GetPath())
            self.text_inputfolder.SetLabel(
                f"Unsorted behavior examples are in: f{self.input_path}."
            )
        dialog.Destroy()

    def output_folder(self, event):
        """Select folder to store sorted behavior examples."""
        dialog = wx.DirDialog(self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE)
        if dialog.ShowModal() == wx.ID_OK:
            self.result_path = Path(dialog.GetPath())
            self.text_outputfolder.SetLabel(
                f"Sorted behavior examples will be in: {self.result_path}."
            )
        dialog.Destroy()

    def parse_key_behavior_pairs(self, key_behavior_pairs: str):
        """Parse a keyboard shortcut behavior string into a dictionary.

        Each key should be associated with a behavior name in the format

            key-behavior, key-behavior, key-behavior...

        where key is a single character corresponding to a key on the keyboard
        and behavior is a string describing the behavior name. The keys 'o',
        'p', 'u', and 'q' (case-insensitive) are reserved for the 'previous',
        'next', 'quit', and 'undo' operations respectively.

        Args:
            key_behavior_pairs: The user-submitted key-behavior pair string.

        Returns:
            None. The shortcuts are stored in the self.keys_behaviors
            attribute.

        Raises:
            ValueError: There was an issue parsing the input string.
        """
        self.keys_behaviors = {}
        RESERVED_KEYS = ["o", "p", "u", "q"]
        for pair in key_behavior_pairs.split(","):
            pair = pair.strip()

            if len(pair.split("-")) != 2:
                raise ValueError(f"Invalid key-behavior pair '{pair}'.")

            key = pair.split("-")[0]
            behavior_name = pair.split("-")[1]

            if len(key) != 1:
                raise ValueError(
                    f"Invalid key '{key}'. Key must be a single character."
                )

            if key.lower() in RESERVED_KEYS:
                raise ValueError(
                    f"Key '{key}' is reserved. Please use a different key."
                )

            self.keys_behaviors[key] = behavior_name

    def input_keys(self, event):
        """Enter keyboard shortcuts to automate sorting."""
        while True:
            dialog = wx.TextEntryDialog(
                self,
                'Enter key-behaviorname pairs separated by ",".',
                "Format: key1-name1,key2-name2,...",
                value="",
            )
            if dialog.ShowModal() != wx.ID_OK:
                dialog.Destroy()
                break
            else:
                key_behavior_pairs = dialog.GetValue()
                dialog.Destroy()

            try:
                self.parse_key_behavior_pairs(key_behavior_pairs)
            except ValueError as err:
                wx.MessageBox(
                    str(err),
                    "Error",
                    wx.OK | wx.ICON_ERROR,
                )
                continue

            self.text_keynames.SetLabel(
                f"The key-behaviorname pairs: {key_behavior_pairs}."
            )
            break

    def _generate_behavior_preview(
        self, pattern_image: MatLike, animation: cv2.VideoCapture | None = None
    ) -> MatLike:
        """Generate the behavior example display to show to the user."""

        def put_example_text(img: MatLike, text: str, pos: Tuple[int, int]):
            """Put the given text on the image at the given position."""
            cv2.putText(
                img,
                text,
                pos,
                cv2.FONT_HERSHEY_SIMPLEX,
                0.5,
                (255, 255, 255),
                1,
            )

        if animation:
            ret, frame = animation.read()
            if not ret:
                animation.set(cv2.CAP_PROP_POS_FRAMES, 0)
                ret, frame = animation.read()

            frame = cv2.resize(frame, (600, 600), interpolation=cv2.INTER_AREA)
            combined = np.hstack((frame, pattern_image))
            put_example_text(
                combined,
                f"frame count: {int(animation.get(cv2.CAP_PROP_FRAME_COUNT))}",
                (10, 15),
            )
            x_begin = 550
        else:
            combined = pattern_image
            x_begin = 5

        # Add keyboard shortcut overlay
        n = 1
        for i in ["o: Prev", "p: Next", "q: Quit", "u: Undo"]:
            put_example_text(combined, i, (x_begin, 15 * n))
            n += 1
        n += 1
        for key, behavior in self.keys_behaviors.items():
            put_example_text(combined, f"{key}: {behavior}", (x_begin, 15 * n))
            n += 1
        return combined

    def sort_behaviors(self, event):
        """Sort behavior examples."""
        if (
            self.input_path is None
            or self.result_path is None
            or not self.keys_behaviors
        ):
            wx.MessageBox(
                "No input / output folder or shortcut key - behavior name pair specified.",
                "Error",
                wx.OK | wx.ICON_ERROR,
            )
            return

        try:
            sorter = BehaviorExampleSorter(
                source=self.input_path,
                destination=self.result_path,
                behaviors=list(self.keys_behaviors.values()),
            )
        except ValueError as err:
            wx.MessageBox(str(err), "Error", wx.OK | wx.ICON_ERROR)
            return

        cv2.namedWindow("Sorting Behavior Examples", cv2.WINDOW_NORMAL)

        moved = False
        stop = False
        example_name = ""
        behavior = ""
        while stop is False:
            if moved:
                sorter.sort_example(example_name, behavior)
                moved = False

            # Check for remaining animations/pattern images
            if sorter.num_remaining_examples == 0:
                wx.MessageBox(
                    "Behavior example sorting completed!",
                    "Completed!",
                    wx.OK | wx.ICON_INFORMATION,
                )
                stop = True
                continue

            # Load the pattern image and animation
            pattern_image = cv2.resize(
                cv2.imread(sorter.current_pattern_image_path),
                (600, 600),
                interpolation=cv2.INTER_AREA,
            )
            if not sorter.IMAGES_ONLY:
                animation = cv2.VideoCapture(sorter.current_animation_path)
                fps = round(animation.get(cv2.CAP_PROP_FPS))

            # This loop is for repeatedly displaying the animation frames
            while True:
                preview = self._generate_behavior_preview(
                    pattern_image,
                    animation if not sorter.IMAGES_ONLY else None,  # pyright: ignore
                )
                cv2.imshow("Sorting Behavior Examples", preview)
                cv2.moveWindow("Sorting Behavior Examples", 50, 0)

                if not sorter.IMAGES_ONLY:
                    pressed = cv2.waitKey(int(1000 / fps)) & 0xFF  # pyright: ignore
                else:
                    pressed = cv2.waitKey(1) & 0xFF

                for key in self.keys_behaviors.keys():
                    if pressed == ord(key):
                        example_name = sorter.current_example
                        behavior = self.keys_behaviors[key]
                        moved = True
                        break
                if moved is True:
                    break

                if pressed == ord("u"):
                    try:
                        sorter.undo()
                    except IndexError as err:
                        wx.MessageBox(str(err), "Error", wx.OK | wx.ICON_ERROR)
                        continue
                    break

                if pressed == ord("p"):
                    sorter.next()
                    break

                if pressed == ord("o"):
                    sorter.prev()
                    break

                if pressed == ord("q"):
                    stop = True
                    break

            if not sorter.IMAGES_ONLY:
                animation.release()  # pyright: ignore

        cv2.destroyAllWindows()


class BehaviorExampleSorter:
    """A class to manage behavior example sorting."""

    def __init__(self, source: Path, destination: Path, behaviors: list[str]) -> None:
        """Initialize a BehaviorExampleSorter.

        Args:
            source: The path to the folder containing unsorted behavior
                examples.
            destination: The path to the folder that will contain sorted
                behavior examples.
            behaviors: A list of the behaviors to sort into.

        Raises:
            ValueError: Unable to find a valid set of behavior examples in the
                source folder.
        """
        self.source = source
        self.destination = destination

        animations = sorted(list(self.source.glob("*.avi")))
        pattern_images = sorted(list(self.source.glob("*.jpg")))
        if len(animations) == 0 and len(pattern_images) == 0:
            raise ValueError("Unable to find behavior examples!")

        animation_example_names = [animation.stem for animation in animations]
        pattern_image_example_names = [image.stem for image in pattern_images]
        if sorted(animation_example_names) != sorted(pattern_image_example_names):
            raise ValueError("Mismatched animations and pattern images!")

        self.IMAGES_ONLY = len(animations) == 0

        self._behaviors = behaviors
        for behavior in self._behaviors:
            (self.destination / behavior).mkdir(exist_ok=True)

        self._action_stack = deque()
        self._index = 0

    @property
    def current_animation_path(self) -> str:
        """The path to the current animation."""
        animations = sorted(list(self.source.glob("*.avi")))
        return str(self.source / animations[self._index])

    @property
    def current_pattern_image_path(self) -> str:
        """The path to the current pattern image."""
        pattern_images = sorted(list(self.source.glob("*.jpg")))
        return str(self.source / pattern_images[self._index])

    @property
    def current_example(self) -> str:
        pattern_images = sorted(list(self.source.glob("*.jpg")))
        return pattern_images[self._index].stem

    @property
    def _remaining_examples(self) -> list[str]:
        """The names of the remaining unsorted behavior examples."""
        return [example.stem for example in self.source.glob("*.jpg")]

    @property
    def num_remaining_examples(self) -> int:
        """The number of remaining unsorted behavior examples."""
        return len(self._remaining_examples)

    def sort_example(self, example: str, behavior: str) -> None:
        """Sort the given behavior example to the folder of the given behavior.

        Args:
            example: The name of the behavior example (without the file
                extension) to sort.
            behavior: The behavior corresponding to the given example.

        Raises:
            ValueError: Invalid behavior name or example name.
        """
        if example not in self._remaining_examples:
            raise ValueError(f"Invalid behavior example {example}.")
        if behavior not in self._behaviors:
            raise ValueError(f"Invalid behavior name {behavior}.")

        (self.source / (example + ".jpg")).rename(
            self.destination / behavior / (example + ".jpg")
        )
        if not self.IMAGES_ONLY:
            (self.source / (example + ".avi")).rename(
                self.destination / behavior / (example + ".avi")
            )
        self._action_stack.append((behavior, example))

    def undo(self):
        """Undo the most recent sorting operation.

        Raises:
            IndexError: No operations left to undo.
        """
        if len(self._action_stack) == 0:
            raise IndexError("Nothing to undo.")

        behavior, example = self._action_stack.pop()
        (self.destination / behavior / (example + ".jpg")).rename(
            self.source / (example + ".jpg")
        )
        if not self.IMAGES_ONLY:
            (self.destination / behavior / example + ".avi").rename(
                self.source / (example + ".avi")
            )

    def next(self):
        """Skip to the next behavior example."""
        self._index = (self._index + 1) % self.num_remaining_examples

    def prev(self):
        """Go to the previous behavior example."""
        self._index = max(0, self._index - 1)



"""
Copyright (C)
This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program. If not, see https://tldrlegal.com/license/gnu-general-public-license-v3-(gpl-3)#fulltext. 

For license issues, please contact:

Dr. Bing Ye
Life Sciences Institute
University of Michigan
210 Washtenaw Avenue, Room 5403
Ann Arbor, MI 48109-2216
USA

Email: bingye@umich.edu
"""


import datetime
import functools
import gc
import math
import operator
import os
from collections import deque

import cv2
import numpy as np
import pandas as pd
import tensorflow as tf
from scipy.spatial import distance
from tensorflow.keras.models import load_model
from tensorflow.keras.preprocessing.image import img_to_array

from .tools import *


class AnalyzeAnimal:
    def __init__(self):
        self.path_to_video = None
        self.basename = None
        self.fps = None
        self.framewidth = None
        self.frameheight = None
        self.kernel = 3
        self.results_path = None
        self.dim_tconv = 8
        self.dim_conv = 8
        self.channel = 1
        self.include_bodyparts = False
        self.std = 0
        self.categorize_behavior = False
        self.animation_analyzer = True
        self.delta = None
        self.animal_number = None
        self.autofind_t = False
        self.t = 0
        self.duration = 5
        self.length = None
        self.animal_area = None
        self.animal_vs_bg = None
        self.background = None
        self.background_low = None
        self.background_high = None
        self.skipped_frames = []
        self.all_time = []
        self.total_analysis_framecount = None
        self.to_deregister = {}
        self.count_to_deregister = None
        self.register_counts = {}
        self.animal_contours = {}
        self.animal_centers = {}
        self.animal_existingcenters = {}
        self.animal_heights = {}
        self.animal_inners = {}
        self.animal_blobs = {}
        self.animations = {}
        self.pattern_images = {}
        self.event_probability = {}
        self.all_behavior_parameters = {}

    def prepare_analysis(
        self,
        path_to_video,
        results_path,
        animal_number,
        delta=1.2,  # an estimated fold change (1.2) of the light intensity for optogenetics only
        names_and_colors=None,  # the behavior names and their representing colors
        framewidth=None,  # the width of the frame to resize
        stable_illumination=True,  # whether the illumination in videos is stable
        dim_tconv=8,  # input dim of Animation Analyzer
        dim_conv=8,  # input dim of Pattern Recognizer
        channel=1,  # color of Animation Analyzer
        include_bodyparts=False,  # whether to include body parts in pattern images
        std=0,  # for excluding static pixels in pattern images
        categorize_behavior=True,  # whether to categorize behaviors or just track animals without behavioral categorization
        animation_analyzer=True,  # whether to include Animation Analyzer
        path_background=None,  # if not None, load backgrounds (need to put the background images under the name 'background.jpg', 'background_low.jpg' and 'background_high.jpg' in the 'path_background' path)
        autofind_t=False,  # whether to use the time point of the first time of illumination change as start_t
        t=0,  # if autofind_t is False, t is the start_t
        duration=5,  # the duration for example generation / analysis
        ex_start=0,  # the start time point for background extraction
        ex_end=None,  # the end time point for background extraction, if None, use the entire video
        length=15,  # the duration (number of frames) of a behavior example (a behavior episode)
        animal_vs_bg=0,  # 0: animals brighter than the background; 1: animals darker than the background; 2: hard to tell
    ):
        print("Preparation started...")
        print(datetime.datetime.now())

        self.path_to_video = path_to_video
        self.basename = os.path.basename(self.path_to_video)
        self.framewidth = framewidth
        self.results_path = os.path.join(
            results_path, os.path.splitext(self.basename)[0]
        )
        self.animal_number = animal_number
        self.dim_tconv = dim_tconv
        self.dim_conv = dim_conv
        self.channel = channel
        self.include_bodyparts = include_bodyparts
        self.std = std
        self.categorize_behavior = categorize_behavior
        self.animation_analyzer = animation_analyzer
        self.autofind_t = autofind_t
        self.t = t
        self.duration = duration
        self.length = length
        os.makedirs(self.results_path, exist_ok=True)
        capture = cv2.VideoCapture(self.path_to_video)
        self.fps = round(capture.get(cv2.CAP_PROP_FPS))
        self.count_to_deregister = self.fps * 2

        if self.duration <= 0:
            self.total_analysis_framecount = (
                int(capture.get(cv2.CAP_PROP_FRAME_COUNT)) + 1
            )
        else:
            self.total_analysis_framecount = int(self.fps * self.duration) + 1

        while True:
            retval, frame = capture.read()
            break
        capture.release()

        print("Video fps: " + str(self.fps))
        print(
            "The original video framesize: "
            + str(int(frame.shape[0]))
            + " X "
            + str(int(frame.shape[1]))
        )

        if self.framewidth is not None:
            self.frameheight = int(frame.shape[0] * self.framewidth / frame.shape[1])
            self.background = cv2.resize(
                frame, (self.framewidth, self.frameheight), interpolation=cv2.INTER_AREA
            )
            print(
                "The resized video framesize: "
                + str(self.frameheight)
                + " X "
                + str(self.framewidth)
            )
        else:
            self.background = frame

        smallest_dimension = min(self.background.shape[0], self.background.shape[1])
        if smallest_dimension / self.animal_number < 250:
            self.kernel = 3
        elif smallest_dimension / self.animal_number < 500:
            self.kernel = 5
        elif smallest_dimension / self.animal_number < 1000:
            self.kernel = 7
        elif smallest_dimension / self.animal_number < 1500:
            self.kernel = 9
        else:
            self.kernel = 11

        self.delta = delta
        self.autofind_t = autofind_t
        self.animal_vs_bg = animal_vs_bg

        if self.autofind_t is True:
            es_start = None
        else:
            es_start = self.t

        if self.framewidth is None or self.frameheight is None:
            frame_size = None
        else:
            frame_size = (self.framewidth, self.frameheight)

        if path_background is None:
            backgrounds = extract_backgrounds_from_video(
                self.path_to_video,
                self.delta,
                frame_size,
                stable_illumination,
                ex_start,
                ex_end,
                self.animal_vs_bg,
            )
        else:
            backgrounds = load_backgrounds_from_folder(path_background, frame_size)

        self.background = backgrounds[0]
        self.background_low = backgrounds[1]
        self.background_high = backgrounds[2]
        cv2.imwrite(os.path.join(self.results_path, "background.jpg"), self.background)
        cv2.imwrite(
            os.path.join(self.results_path, "background_low.jpg"), self.background_low
        )
        cv2.imwrite(
            os.path.join(self.results_path, "background_high.jpg"), self.background_high
        )

        if self.autofind_t:
            self.t = get_stimulation_time(self.path_to_video, self.delta)

        self.animal_area = estimate_animal_area(
            self.path_to_video,
            self.delta,
            self.background,
            self.background_low,
            self.background_high,
            self.animal_number,
            frame_size,
            self.t,
            es_start,
            self.duration,
            self.animal_vs_bg,
            self.kernel,
        )

        if self.categorize_behavior is True:
            for behavior_name in names_and_colors:
                self.all_behavior_parameters[behavior_name] = {}
                self.all_behavior_parameters[behavior_name]["color"] = names_and_colors[
                    behavior_name
                ]
                for parameter_name in [
                    "acceleration",
                    "count",
                    "distance",
                    "duration",
                    "intensity_area",
                    "intensity_length",
                    "latency",
                    "magnitude_area",
                    "magnitude_length",
                    "probability",
                    "speed",
                    "velocity",
                    "vigor_area",
                    "vigor_length",
                ]:
                    self.all_behavior_parameters[behavior_name][parameter_name] = {}
        else:
            self.dim_conv = 8
            self.animation_analyzer = False
            for parameter_name in [
                "acceleration",
                "distance",
                "intensity_area",
                "intensity_length",
                "magnitude_area",
                "magnitude_length",
                "speed",
                "velocity",
                "vigor_area",
                "vigor_length",
            ]:
                self.all_behavior_parameters[parameter_name] = {}

        for i in range(self.animal_number):
            self.to_deregister[i] = 0
            self.register_counts[i] = None
            self.animal_contours[i] = [None] * self.total_analysis_framecount
            self.animal_centers[i] = [None] * self.total_analysis_framecount
            self.animal_existingcenters[i] = (-10000, -10000)
            self.animal_heights[i] = [None] * self.total_analysis_framecount
            if self.include_bodyparts is True:
                self.animal_inners[i] = deque(maxlen=self.length)
            if self.animation_analyzer is True:
                self.animal_blobs[i] = (
                    deque(
                        [
                            np.zeros(
                                (self.dim_tconv, self.dim_tconv, self.channel),
                                dtype="uint8",
                            )
                        ],
                        maxlen=self.length,
                    )
                    * self.length
                )
                self.animations[i] = [
                    np.zeros(
                        (self.length, self.dim_tconv, self.dim_tconv, self.channel),
                        dtype="uint8",
                    )
                ] * self.total_analysis_framecount
            self.pattern_images[i] = [
                np.zeros((self.dim_conv, self.dim_conv, 3), dtype="uint8")
            ] * self.total_analysis_framecount

        print("Preparation completed!")

    def track_animal(
        self, frame_count_analyze, contours, centers, heights, inners=None, blobs=None
    ):
        unused_existing_indices = list(self.animal_existingcenters)
        existing_centers = list(self.animal_existingcenters.values())
        unused_new_indices = list(range(len(centers)))
        dt_flattened = distance.cdist(existing_centers, centers).flatten()
        dt_sort_index = dt_flattened.argsort()
        length = len(centers)

        for idx in dt_sort_index:
            index_in_existing = int(idx / length)
            index_in_new = int(idx % length)
            if index_in_existing in unused_existing_indices:
                if index_in_new in unused_new_indices:
                    unused_existing_indices.remove(index_in_existing)
                    unused_new_indices.remove(index_in_new)
                    if self.register_counts[index_in_existing] is None:
                        self.register_counts[index_in_existing] = frame_count_analyze
                    self.to_deregister[index_in_existing] = 0
                    self.animal_contours[index_in_existing][
                        frame_count_analyze
                    ] = contours[index_in_new]
                    center = centers[index_in_new]
                    self.animal_centers[index_in_existing][frame_count_analyze] = center
                    self.animal_existingcenters[index_in_existing] = center
                    self.animal_heights[index_in_existing][
                        frame_count_analyze
                    ] = heights[index_in_new]
                    if self.animation_analyzer is True:
                        blob = blobs[index_in_new]
                        blob = img_to_array(
                            cv2.resize(
                                blob,
                                (self.dim_tconv, self.dim_tconv),
                                interpolation=cv2.INTER_AREA,
                            )
                        )
                        self.animal_blobs[index_in_existing].append(blob)
                        self.animations[index_in_existing][
                            frame_count_analyze
                        ] = np.array(self.animal_blobs[index_in_existing])
                    if self.include_bodyparts is True:
                        self.animal_inners[index_in_existing].append(
                            inners[index_in_new]
                        )
                        pattern_image = generate_patternimage(
                            self.background,
                            self.animal_contours[index_in_existing][
                                max(
                                    0, (frame_count_analyze - self.length + 1)
                                ) : frame_count_analyze + 1
                            ],
                            inners=self.animal_inners[index_in_existing],
                            std=self.std,
                        )
                    else:
                        pattern_image = generate_patternimage(
                            self.background,
                            self.animal_contours[index_in_existing][
                                max(
                                    0, (frame_count_analyze - self.length + 1)
                                ) : frame_count_analyze + 1
                            ],
                            inners=None,
                            std=0,
                        )
                    pattern_image = cv2.resize(
                        pattern_image,
                        (self.dim_conv, self.dim_conv),
                        interpolation=cv2.INTER_AREA,
                    )
                    self.pattern_images[index_in_existing][
                        frame_count_analyze
                    ] = np.array(pattern_image)

        if len(unused_existing_indices) > 0:
            for i in unused_existing_indices:
                if self.to_deregister[i] <= self.count_to_deregister:
                    self.to_deregister[i] += 1
                else:
                    self.animal_existingcenters[i] = (-10000, -10000)
                if self.animation_analyzer is True:
                    self.animal_blobs[i].append(
                        np.zeros(
                            (self.dim_tconv, self.dim_tconv, self.channel),
                            dtype="uint8",
                        )
                    )
                if self.include_bodyparts is True:
                    self.animal_inners[i].append(None)

    def acquire_information(self, background_free=True):
        print("Acquiring information in each frame...")
        print(datetime.datetime.now())

        capture = cv2.VideoCapture(self.path_to_video)

        background = self.background
        background_low = self.background_low
        background_high = self.background_high
        if self.animal_vs_bg == 1:
            background = np.uint8(255 - background)
            background_low = np.uint8(255 - background_low)
            background_high = np.uint8(255 - background_high)

        frame_count = frame_count_analyze = 0
        if background_free is False:
            temp_frames = deque(maxlen=self.length)
            animation = (
                deque(
                    [
                        np.zeros(
                            (self.dim_tconv, self.dim_tconv, self.channel),
                            dtype="uint8",
                        )
                    ],
                    maxlen=self.length,
                )
                * self.length
            )

        start_t = round((self.t - self.length / self.fps), 2)
        if start_t < 0:
            start_t = 0.00
        if self.duration == 0:
            end_t = float("inf")
        else:
            end_t = start_t + self.duration

        while True:
            retval, frame = capture.read()
            time = round((frame_count + 1) / self.fps, 2)

            if time >= end_t or frame is None:
                break

            if time >= start_t:
                self.all_time.append(round((time - start_t), 2))

                if (frame_count_analyze + 1) % 1000 == 0:
                    print(str(frame_count_analyze + 1) + " frames processed...")
                    print(datetime.datetime.now())

                if self.framewidth is not None:
                    frame = cv2.resize(
                        frame,
                        (self.framewidth, self.frameheight),
                        interpolation=cv2.INTER_AREA,
                    )

                if background_free is False:
                    temp_frames.append(frame)

                (contours, centers, heights, inners, blobs) = contour_frame(
                    frame,
                    self.animal_number,
                    background,
                    background_low,
                    background_high,
                    self.delta,
                    self.animal_area,
                    animal_vs_bg=self.animal_vs_bg,
                    include_bodyparts=self.include_bodyparts,
                    animation_analyzer=self.animation_analyzer,
                    channel=self.channel,
                    kernel=self.kernel,
                )

                if len(contours) == 0:
                    self.skipped_frames.append(frame_count_analyze)

                    for i in self.animal_centers:
                        if self.animation_analyzer is True:
                            self.animal_blobs[i].append(
                                np.zeros(
                                    (self.dim_tconv, self.dim_tconv, self.channel),
                                    dtype="uint8",
                                )
                            )
                        if self.include_bodyparts is True:
                            self.animal_inners[i].append(None)

                else:
                    self.track_animal(
                        frame_count_analyze,
                        contours,
                        centers,
                        heights,
                        inners=inners,
                        blobs=blobs,
                    )

                    if self.animation_analyzer is True:
                        if background_free is False:
                            for i in self.animal_centers:
                                for n, f in enumerate(temp_frames):
                                    if (
                                        self.animal_contours[i][
                                            max(
                                                0,
                                                (frame_count_analyze - self.length + 1),
                                            ) : frame_count_analyze + 1
                                        ][n]
                                        is None
                                    ):
                                        blob = np.zeros(
                                            (
                                                self.dim_tconv,
                                                self.dim_tconv,
                                                self.channel,
                                            ),
                                            dtype="uint8",
                                        )
                                    else:
                                        blob = extract_blob_background(
                                            f,
                                            self.animal_contours[i][
                                                max(
                                                    0,
                                                    (
                                                        frame_count_analyze
                                                        - self.length
                                                        + 1
                                                    ),
                                                ) : frame_count_analyze + 1
                                            ],
                                            contour=None,
                                            channel=self.channel,
                                            background_free=False,
                                        )
                                        blob = cv2.resize(
                                            blob,
                                            (self.dim_tconv, self.dim_tconv),
                                            interpolation=cv2.INTER_AREA,
                                        )
                                    animation.append(img_to_array(blob))
                                self.animations[i][frame_count_analyze] = np.array(
                                    animation
                                )

                frame_count_analyze += 1

            frame_count += 1

        capture.release()

        print("Information acquisition completed!")

    def acquire_information_interact_basic(self, background_free=True):
        print("Acquiring information in each frame...")
        print(datetime.datetime.now())

        if self.animation_analyzer is True:
            self.animations[0] = [
                np.zeros(
                    (self.length, self.dim_tconv, self.dim_tconv, self.channel),
                    dtype="uint8",
                )
            ] * self.total_analysis_framecount
        self.pattern_images[0] = [
            np.zeros((self.dim_conv, self.dim_conv, 3), dtype="uint8")
        ] * self.total_analysis_framecount
        self.animal_contours[0] = [None] * self.total_analysis_framecount
        self.animal_centers[0] = [None] * self.total_analysis_framecount

        background = self.background
        background_low = self.background_low
        background_high = self.background_high
        if self.animal_vs_bg == 1:
            background = np.uint8(255 - background)
            background_low = np.uint8(255 - background_low)
            background_high = np.uint8(255 - background_high)

        capture = cv2.VideoCapture(self.path_to_video)
        frame_count = frame_count_analyze = 0
        temp_frames = deque(maxlen=self.length)
        temp_contours = deque(maxlen=self.length)
        temp_inners = deque(maxlen=self.length)
        animation = (
            deque(
                [
                    np.zeros(
                        (self.dim_tconv, self.dim_tconv, self.channel), dtype="uint8"
                    )
                ],
                maxlen=self.length,
            )
            * self.length
        )

        start_t = round((self.t - self.length / self.fps), 2)
        if start_t < 0:
            start_t = 0.00
        if self.duration == 0:
            end_t = float("inf")
        else:
            end_t = start_t + self.duration

        while True:
            retval, frame = capture.read()
            time = round((frame_count + 1) / self.fps, 2)

            if time >= end_t or frame is None:
                break

            if time >= start_t:
                self.all_time.append(round((time - start_t), 2))

                if (frame_count_analyze + 1) % 1000 == 0:
                    print(str(frame_count_analyze + 1) + " frames processed...")
                    print(datetime.datetime.now())

                if self.framewidth is not None:
                    frame = cv2.resize(
                        frame,
                        (self.framewidth, self.frameheight),
                        interpolation=cv2.INTER_AREA,
                    )

                temp_frames.append(frame)

                (contours, centers, heights, inners, blobs) = contour_frame(
                    frame,
                    self.animal_number,
                    background,
                    background_low,
                    background_high,
                    self.delta,
                    self.animal_area,
                    animal_vs_bg=self.animal_vs_bg,
                    include_bodyparts=self.include_bodyparts,
                    animation_analyzer=False,
                    channel=self.channel,
                    kernel=self.kernel,
                )

                if len(contours) == 0:
                    self.skipped_frames.append(frame_count_analyze)

                    temp_contours.append(None)
                    temp_inners.append(None)
                    animation.append(
                        np.zeros(
                            (self.dim_tconv, self.dim_tconv, self.channel),
                            dtype="uint8",
                        )
                    )

                else:
                    temp_contours.append(contours)
                    temp_inners.append(inners)
                    length = len(temp_frames)
                    if length == 1:
                        self.register_counts[0] = frame_count_analyze
                    self.animal_contours[0][frame_count_analyze] = contours

                    (y_bt, y_tp, x_lf, x_rt) = crop_frame(
                        frame, functools.reduce(operator.iconcat, temp_contours, [])
                    )
                    self.animal_centers[0][frame_count_analyze] = (x_lf + 5, y_bt + 10)

                    pattern_image = generate_patternimage_all(
                        frame,
                        y_bt,
                        y_tp,
                        x_lf,
                        x_rt,
                        temp_contours,
                        temp_inners,
                        std=self.std,
                    )
                    self.pattern_images[0][frame_count_analyze] = np.array(
                        cv2.resize(
                            pattern_image,
                            (self.dim_conv, self.dim_conv),
                            interpolation=cv2.INTER_AREA,
                        )
                    )

                    if self.animation_analyzer is True:
                        for i, f in enumerate(temp_frames):
                            if (
                                self.animal_contours[0][
                                    max(
                                        0, (frame_count_analyze - self.length + 1)
                                    ) : frame_count_analyze + 1
                                ][n]
                                is None
                            ):
                                blob = np.zeros(
                                    (self.dim_tconv, self.dim_tconv, self.channel),
                                    dtype="uint8",
                                )
                            else:
                                blob = extract_blob_all(
                                    f,
                                    y_bt,
                                    y_tp,
                                    x_lf,
                                    x_rt,
                                    contours=temp_contours[i],
                                    channel=self.channel,
                                    background_free=background_free,
                                )
                            blob = cv2.resize(
                                blob,
                                (self.dim_tconv, self.dim_tconv),
                                interpolation=cv2.INTER_AREA,
                            )
                            animation.append(img_to_array(blob))
                            self.animations[0][frame_count_analyze] = np.array(
                                animation
                            )

                frame_count_analyze += 1

            frame_count += 1

        capture.release()

        self.animations[0] = self.animations[0][: len(self.all_time)]
        self.pattern_images[0] = self.pattern_images[0][: len(self.all_time)]
        self.animal_contours[0] = self.animal_contours[0][: len(self.all_time)]
        self.animal_centers[0] = self.animal_centers[0][: len(self.all_time)]

        print("Information acquisition completed!")

    def craft_data(self):
        print("Crafting data...")
        print(datetime.datetime.now())

        lengths = []
        length = len(self.all_time)
        to_delete = []
        IDs = list(self.animal_centers.keys())

        for i in IDs:
            check = 0
            for x in self.animal_heights[i]:
                if x is not None:
                    check += 1
            lengths.append(check)
            if self.register_counts[i] is None:
                to_delete.append(i)

        if len(IDs) == len(to_delete):
            to_delete.remove(np.argsort(lengths)[-1])

        for i in IDs:
            if i in to_delete:
                del self.to_deregister[i]
                del self.register_counts[i]
                del self.animal_centers[i]
                del self.animal_existingcenters[i]
                del self.animal_contours[i]
                del self.animal_heights[i]
                if self.include_bodyparts is True:
                    del self.animal_inners[i]
                if self.animation_analyzer is True:
                    del self.animal_blobs[i]
                    del self.animations[i]
                del self.pattern_images[i]

        for i in self.animal_centers:
            self.animal_centers[i] = self.animal_centers[i][:length]
            self.animal_contours[i] = self.animal_contours[i][:length]
            self.animal_heights[i] = self.animal_heights[i][:length]
            if self.animation_analyzer is True:
                self.animations[i] = self.animations[i][:length]
            self.pattern_images[i] = self.pattern_images[i][:length]

        print("Data crafting completed!")

    def categorize_behaviors(self, path_to_categorizer, uncertain=0):
        print("Categorizing behaviors...")
        print(datetime.datetime.now())

        IDs = list(self.pattern_images.keys())

        if self.animation_analyzer is True:
            animations = self.animations[IDs[0]]
        pattern_images = self.pattern_images[IDs[0]]

        if len(self.pattern_images) > 1:
            for n in IDs[1:]:
                if self.animation_analyzer is True:
                    animations += self.animations[n]
                pattern_images += self.pattern_images[n]

        del self.animations
        del self.pattern_images
        gc.collect()

        with tf.device("CPU"):
            if self.animation_analyzer is True:
                animations = tf.convert_to_tensor(
                    np.array(animations, dtype="float32") / 255.0
                )
            pattern_images = tf.convert_to_tensor(
                np.array(pattern_images, dtype="float32") / 255.0
            )

        if self.animation_analyzer is True:
            inputs = [animations, pattern_images]
        else:
            inputs = pattern_images

        categorizer = load_model(path_to_categorizer)
        predictions = categorizer.predict(inputs, batch_size=32)

        for behavior_name in self.all_behavior_parameters:
            for i in IDs:
                self.all_behavior_parameters[behavior_name]["probability"][i] = [
                    np.nan
                ] * len(self.all_time)
                self.event_probability[i] = [["NA", -1]] * len(self.all_time)

        idx = 0
        for n in IDs:
            i = self.length + self.register_counts[n]
            idx += i
            while i < len(self.animal_contours[n]):
                if self.animal_contours[n][i] is not None:
                    check = 0
                    for c in self.animal_contours[n][i - self.length + 1 : i + 1]:
                        if c is None:
                            check += 1
                    if check <= self.length / 2:
                        prediction = predictions[idx]
                        behavior_names = list(self.all_behavior_parameters.keys())
                        for behavior_name in behavior_names:
                            if len(behavior_names) == 2:
                                if behavior_names.index(behavior_name) == 0:
                                    probability = 1 - prediction[0]
                                else:
                                    probability = prediction[0]
                            else:
                                probability = prediction[
                                    behavior_names.index(behavior_name)
                                ]
                            self.all_behavior_parameters[behavior_name]["probability"][
                                n
                            ][i] = probability
                        if len(behavior_names) == 2:
                            if prediction[0] > 0.5:
                                if prediction[0] - (1 - prediction[0]) > uncertain:
                                    self.event_probability[n][i] = [
                                        behavior_names[1],
                                        prediction[0],
                                    ]
                            if prediction[0] < 0.5:
                                if (1 - prediction[0]) - prediction[0] > uncertain:
                                    self.event_probability[n][i] = [
                                        behavior_names[0],
                                        1 - prediction[0],
                                    ]
                        else:
                            if (
                                sorted(prediction)[-1] - sorted(prediction)[-2]
                                > uncertain
                            ):
                                self.event_probability[n][i] = [
                                    behavior_names[np.argmax(prediction)],
                                    max(prediction),
                                ]

                idx += 1
                i += 1

        del predictions
        gc.collect()

        print("Behavioral categorization completed!")

    def annotate_video(self, behavior_to_include, show_legend=True, interact_all=False):
        print("Annotating video...")
        print(datetime.datetime.now())

        text_scl = max(
            min(self.background.shape[0], self.background.shape[1]) / 960, 0.5
        )
        text_tk = max(
            1, int(min(self.background.shape[0], self.background.shape[1]) / 960)
        )

        if self.categorize_behavior is True:
            colors = {}
            for behavior_name in self.all_behavior_parameters:
                if self.all_behavior_parameters[behavior_name]["color"][1][0] != "#":
                    colors[behavior_name] = (255, 255, 255)
                else:
                    hex_color = self.all_behavior_parameters[behavior_name]["color"][
                        1
                    ].lstrip("#")
                    color = tuple(int(hex_color[i : i + 2], 16) for i in (0, 2, 4))
                    colors[behavior_name] = color[::-1]

            if len(behavior_to_include) != len(self.all_behavior_parameters):
                for behavior_name in self.all_behavior_parameters:
                    if behavior_name not in behavior_to_include:
                        del colors[behavior_name]

            if show_legend is True:
                scl = self.background.shape[0] / 1024
                if 25 * (len(colors) + 1) < self.background.shape[0]:
                    intvl = 25
                else:
                    intvl = int(self.background.shape[0] / (len(colors) + 1))

        capture = cv2.VideoCapture(self.path_to_video)
        writer = None
        frame_count = frame_count_analyze = index = 0

        start_t = round((self.t - self.length / self.fps), 2)
        if start_t < 0:
            start_t = 0.00
        if self.duration == 0:
            end_t = float("inf")
        else:
            end_t = start_t + self.duration

        while True:
            retval, frame = capture.read()
            time = round((frame_count + 1) / self.fps, 2)

            if time >= end_t or frame is None:
                break

            if time >= start_t:
                if self.framewidth is not None:
                    frame = cv2.resize(
                        frame,
                        (self.framewidth, self.frameheight),
                        interpolation=cv2.INTER_AREA,
                    )

                if self.categorize_behavior is True:
                    if show_legend is True:
                        n = 1
                        for i in colors:
                            cv2.putText(
                                frame,
                                i,
                                (10, intvl * n),
                                cv2.FONT_HERSHEY_SIMPLEX,
                                scl,
                                colors[i],
                                text_tk,
                            )
                            n += 1

                if frame_count_analyze not in self.skipped_frames:
                    for i in self.animal_contours:
                        if frame_count_analyze < len(self.animal_contours[i]):
                            if self.animal_contours[i][frame_count_analyze] is not None:
                                cx = self.animal_centers[i][frame_count_analyze][0]
                                cy = self.animal_centers[i][frame_count_analyze][1]

                                if interact_all is False:
                                    cv2.putText(
                                        frame,
                                        str(i),
                                        (cx - 10, cy - 10),
                                        cv2.FONT_HERSHEY_SIMPLEX,
                                        text_scl,
                                        (255, 255, 255),
                                        text_tk,
                                    )
                                    cv2.circle(
                                        frame,
                                        (cx, cy),
                                        int(text_tk * 3),
                                        (255, 0, 0),
                                        -1,
                                    )

                                if self.categorize_behavior is True:
                                    if (
                                        self.event_probability[i][frame_count_analyze][
                                            0
                                        ]
                                        == "NA"
                                    ):
                                        if interact_all is True:
                                            cv2.drawContours(
                                                frame,
                                                self.animal_contours[i][
                                                    frame_count_analyze
                                                ],
                                                -1,
                                                (255, 255, 255),
                                                1,
                                            )
                                        else:
                                            cv2.drawContours(
                                                frame,
                                                [
                                                    self.animal_contours[i][
                                                        frame_count_analyze
                                                    ]
                                                ],
                                                0,
                                                (255, 255, 255),
                                                1,
                                            )
                                        cv2.putText(
                                            frame,
                                            "NA",
                                            (cx + 10, cy - 10),
                                            cv2.FONT_HERSHEY_SIMPLEX,
                                            text_scl,
                                            (255, 255, 255),
                                            text_tk,
                                        )
                                    else:
                                        name = self.event_probability[i][
                                            frame_count_analyze
                                        ][0]
                                        probability = (
                                            str(
                                                round(
                                                    self.event_probability[i][
                                                        frame_count_analyze
                                                    ][1]
                                                    * 100
                                                )
                                            )
                                            + "%"
                                        )
                                        if name in colors:
                                            color = colors[
                                                self.event_probability[i][
                                                    frame_count_analyze
                                                ][0]
                                            ]
                                            if interact_all is True:
                                                cv2.drawContours(
                                                    frame,
                                                    self.animal_contours[i][
                                                        frame_count_analyze
                                                    ],
                                                    -1,
                                                    color,
                                                    1,
                                                )
                                            else:
                                                cv2.drawContours(
                                                    frame,
                                                    [
                                                        self.animal_contours[i][
                                                            frame_count_analyze
                                                        ]
                                                    ],
                                                    0,
                                                    color,
                                                    1,
                                                )
                                            cv2.putText(
                                                frame,
                                                name + " " + probability,
                                                (cx + 10, cy - 10),
                                                cv2.FONT_HERSHEY_SIMPLEX,
                                                text_scl,
                                                color,
                                                text_tk,
                                            )
                                        else:
                                            if interact_all is True:
                                                cv2.drawContours(
                                                    frame,
                                                    self.animal_contours[i][
                                                        frame_count_analyze
                                                    ],
                                                    -1,
                                                    (255, 255, 255),
                                                    1,
                                                )
                                            else:
                                                cv2.drawContours(
                                                    frame,
                                                    [
                                                        self.animal_contours[i][
                                                            frame_count_analyze
                                                        ]
                                                    ],
                                                    0,
                                                    (255, 255, 255),
                                                    1,
                                                )
                                            cv2.putText(
                                                frame,
                                                "NA",
                                                (cx + 10, cy - 10),
                                                cv2.FONT_HERSHEY_SIMPLEX,
                                                text_scl,
                                                (255, 255, 255),
                                                text_tk,
                                            )
                                else:
                                    cv2.drawContours(
                                        frame,
                                        [self.animal_contours[i][frame_count_analyze]],
                                        0,
                                        (255, 255, 255),
                                        1,
                                    )

                    index += 1

                if writer is None:
                    (h, w) = frame.shape[:2]
                    writer = cv2.VideoWriter(
                        os.path.join(self.results_path, "Annotated video.avi"),
                        cv2.VideoWriter_fourcc(*"MJPG"),
                        self.fps,
                        (w, h),
                        True,
                    )

                writer.write(frame)

                frame_count_analyze += 1

            frame_count += 1

        capture.release()
        writer.release()

        print("Video annotation completed!")

    def analyze_parameters(self, normalize_distance=True, parameter_to_analyze=[]):
        all_parameters = []
        if "3 length parameters" in parameter_to_analyze:
            all_parameters += ["intensity_length", "magnitude_length", "vigor_length"]
        if "3 areal parameters" in parameter_to_analyze:
            all_parameters += ["intensity_area", "magnitude_area", "vigor_area"]
        if "4 locomotion parameters" in parameter_to_analyze:
            all_parameters += ["acceleration", "speed", "velocity"]

        if self.categorize_behavior is True:
            for behavior_name in self.all_behavior_parameters:
                for i in self.event_probability:
                    if "count" in parameter_to_analyze:
                        self.all_behavior_parameters[behavior_name]["count"][i] = 0
                    if "duration" in parameter_to_analyze:
                        self.all_behavior_parameters[behavior_name]["duration"][i] = 0.0
                    if "4 locomotion parameters" in parameter_to_analyze:
                        self.all_behavior_parameters[behavior_name]["distance"][i] = 0.0
                    if "latency" in parameter_to_analyze:
                        self.all_behavior_parameters[behavior_name]["latency"][i] = "NA"

                for parameter_name in all_parameters:
                    for i in self.event_probability:
                        self.all_behavior_parameters[behavior_name][parameter_name][
                            i
                        ] = [np.nan] * len(self.all_time)

        else:
            for i in self.animal_contours:
                if "4 locomotion parameters" in parameter_to_analyze:
                    self.all_behavior_parameters["distance"][i] = 0.0
                for parameter_name in all_parameters:
                    self.all_behavior_parameters[parameter_name][i] = [np.nan] * len(
                        self.all_time
                    )

        if len(parameter_to_analyze) > 0:
            for i in self.animal_contours:
                n = self.length + self.register_counts[i]

                while n < len(self.animal_contours[i]):
                    if self.categorize_behavior is True:
                        behavior_name = self.event_probability[i][n][0]

                        if behavior_name != "NA":
                            if "count" in parameter_to_analyze:
                                if behavior_name != self.event_probability[i][n - 1][0]:
                                    self.all_behavior_parameters[behavior_name][
                                        "count"
                                    ][i] += 1

                            if "duration" in parameter_to_analyze:
                                self.all_behavior_parameters[behavior_name]["duration"][
                                    i
                                ] += round(1 / self.fps, 2)

                            if "latency" in parameter_to_analyze:
                                if (
                                    self.all_behavior_parameters[behavior_name][
                                        "latency"
                                    ][i]
                                    == "NA"
                                ):
                                    self.all_behavior_parameters[behavior_name][
                                        "latency"
                                    ][i] = self.all_time[n]

                            if "3 length parameters" in parameter_to_analyze:
                                heights_diffs = []
                                for h in self.animal_heights[i][
                                    n - self.length + 1 : n + 1
                                ]:
                                    if h is None or self.animal_heights[i][n] is None:
                                        height_diff = 0.0
                                    else:
                                        height_diff = (
                                            abs(h - self.animal_heights[i][n]) / h
                                        )
                                    heights_diffs.append(height_diff)
                                magnitude_length = max(heights_diffs)
                                vigor_length = magnitude_length / (
                                    (self.length - np.argmax(heights_diffs)) / self.fps
                                )
                                intensity_length = sum(heights_diffs) / (
                                    self.length / self.fps
                                )
                                if magnitude_length > 0:
                                    self.all_behavior_parameters[behavior_name][
                                        "magnitude_length"
                                    ][i][n] = magnitude_length
                                if vigor_length > 0:
                                    self.all_behavior_parameters[behavior_name][
                                        "vigor_length"
                                    ][i][n] = vigor_length
                                if intensity_length > 0:
                                    self.all_behavior_parameters[behavior_name][
                                        "intensity_length"
                                    ][i][n] = intensity_length

                            if "4 locomotion parameters" in parameter_to_analyze:
                                distance_traveled = 0.0
                                d = n - self.length
                                while d < n - 1:
                                    start_center = self.animal_centers[i][d]
                                    end_center = self.animal_centers[i][d + 1]
                                    if start_center is None or end_center is None:
                                        dt = 0.0
                                    else:
                                        dt = math.dist(end_center, start_center)
                                    distance_traveled += dt
                                    d += 1
                                if normalize_distance is True:
                                    calibrator = math.sqrt(self.animal_area)
                                    distance_traveled = distance_traveled / calibrator
                                speed = distance_traveled / (self.length / self.fps)
                                end_center = self.animal_centers[i][n]
                                if end_center is not None:
                                    displacements = []
                                    for ct in self.animal_centers[i][
                                        n - self.length + 1 : n + 1
                                    ]:
                                        if ct is None:
                                            displacements.append(0)
                                        else:
                                            displacements.append(
                                                math.dist(end_center, ct)
                                            )
                                    displacement = max(displacements)
                                    if normalize_distance is True:
                                        displacement = displacement / calibrator
                                    velocity = displacement / (
                                        (self.length - np.argmax(displacements))
                                        / self.fps
                                    )
                                    self.all_behavior_parameters[behavior_name][
                                        "velocity"
                                    ][i][n] = velocity
                                velocities_max = []
                                velocities_min = []
                                for v in self.all_behavior_parameters[behavior_name][
                                    "velocity"
                                ][i][n - self.length + 1 : n + 1]:
                                    if v != np.nan:
                                        velocities_max.append(v)
                                        velocities_min.append(v)
                                    else:
                                        velocities_max.append(-float("inf"))
                                        velocities_min.append(float("inf"))
                                vmax = max(velocities_max)
                                vmin = min(velocities_min)
                                if vmax != -float("inf") and vmin != float("inf"):
                                    if np.argmax(velocities_max) != np.argmin(
                                        velocities_min
                                    ):
                                        t = (
                                            abs(
                                                np.argmax(velocities_max)
                                                - np.argmin(velocities_min)
                                            )
                                            / self.fps
                                        )
                                        self.all_behavior_parameters[behavior_name][
                                            "acceleration"
                                        ][i][n] = (vmax - vmin) / t
                                self.all_behavior_parameters[behavior_name]["distance"][
                                    i
                                ] += distance_traveled
                                self.all_behavior_parameters[behavior_name]["speed"][i][
                                    n
                                ] = speed

                            if "3 areal parameters" in parameter_to_analyze:
                                mask = np.zeros_like(self.background)
                                contour = self.animal_contours[i][n]
                                if contour is not None:
                                    cv2.drawContours(
                                        mask, [contour], 0, (255, 255, 255), -1
                                    )
                                    mask = cv2.cvtColor(mask, cv2.COLOR_BGR2GRAY)
                                    area_diffs = []
                                    for ct in self.animal_contours[i][
                                        n - self.length + 1 : n + 1
                                    ]:
                                        if ct is not None:
                                            prev_mask = np.zeros_like(self.background)
                                            cv2.drawContours(
                                                prev_mask, [ct], 0, (255, 255, 255), -1
                                            )
                                            prev_mask = cv2.cvtColor(
                                                prev_mask, cv2.COLOR_BGR2GRAY
                                            )
                                            xlf1, ybt1, w, h = cv2.boundingRect(contour)
                                            xrt1 = xlf1 + w
                                            ytp1 = ybt1 + h
                                            xlf2, ybt2, w, h = cv2.boundingRect(ct)
                                            xrt2 = xlf2 + w
                                            ytp2 = ybt2 + h
                                            xlf = min(xlf1, xlf2)
                                            xrt = max(xrt1, xrt2)
                                            ybt = min(ybt1, ybt2)
                                            ytp = max(ytp1, ytp2)
                                            curr_mask = mask[ybt:ytp, xlf:xrt]
                                            prev_mask = prev_mask[ybt:ytp, xlf:xrt]
                                            diff = cv2.bitwise_xor(prev_mask, curr_mask)
                                            # diff[diff!=0]=255
                                            area_diff = (np.sum(diff) / 255) / (
                                                np.sum(prev_mask) / 255
                                            )
                                            area_diffs.append(area_diff)
                                    if len(area_diffs) > 0:
                                        magnitude_area = max(area_diffs)
                                        vigor_area = magnitude_area / (
                                            (self.length - np.argmax(area_diffs))
                                            / self.fps
                                        )
                                        intensity_area = sum(area_diffs) / (
                                            self.length / self.fps
                                        )
                                        if magnitude_area > 0:
                                            self.all_behavior_parameters[behavior_name][
                                                "magnitude_area"
                                            ][i][n] = magnitude_area
                                        if vigor_area > 0:
                                            self.all_behavior_parameters[behavior_name][
                                                "vigor_area"
                                            ][i][n] = vigor_area
                                        if intensity_area > 0:
                                            self.all_behavior_parameters[behavior_name][
                                                "intensity_area"
                                            ][i][n] = intensity_area

                    else:
                        if "3 length parameters" in parameter_to_analyze:
                            heights_diffs = []
                            for h in self.animal_heights[i][
                                n - self.length + 1 : n + 1
                            ]:
                                if h is None or self.animal_heights[i][n] is None:
                                    height_diff = 0.0
                                else:
                                    height_diff = abs(h - self.animal_heights[i][n]) / h
                                heights_diffs.append(height_diff)
                            magnitude_length = max(heights_diffs)
                            vigor_length = magnitude_length / (
                                (self.length - np.argmax(heights_diffs)) / self.fps
                            )
                            intensity_length = sum(heights_diffs) / (
                                self.length / self.fps
                            )
                            if magnitude_length > 0:
                                self.all_behavior_parameters["magnitude_length"][i][
                                    n
                                ] = magnitude_length
                            if vigor_length > 0:
                                self.all_behavior_parameters["vigor_length"][i][
                                    n
                                ] = vigor_length
                            if intensity_length > 0:
                                self.all_behavior_parameters["intensity_length"][i][
                                    n
                                ] = intensity_length

                        if "4 locomotion parameters" in parameter_to_analyze:
                            distance_traveled = 0.0
                            d = n - self.length
                            while d < n - 1:
                                start_center = self.animal_centers[i][d]
                                end_center = self.animal_centers[i][d + 1]
                                if start_center is None or end_center is None:
                                    dt = 0.0
                                else:
                                    dt = math.dist(end_center, start_center)
                                distance_traveled += dt
                                d += 1
                            if normalize_distance is True:
                                calibrator = math.sqrt(self.animal_area)
                                distance_traveled = distance_traveled / calibrator
                            speed = distance_traveled / (self.length / self.fps)
                            end_center = self.animal_centers[i][n]
                            if end_center is not None:
                                displacements = []
                                for ct in self.animal_centers[i][
                                    n - self.length + 1 : n + 1
                                ]:
                                    if ct is None:
                                        displacements.append(0)
                                    else:
                                        displacements.append(math.dist(end_center, ct))
                                displacement = max(displacements)
                                if normalize_distance is True:
                                    displacement = displacement / calibrator
                                velocity = displacement / (
                                    (self.length - np.argmax(displacements)) / self.fps
                                )
                                self.all_behavior_parameters["velocity"][i][
                                    n
                                ] = velocity
                            velocities_max = []
                            velocities_min = []
                            for v in self.all_behavior_parameters["velocity"][i][
                                n - self.length + 1 : n + 1
                            ]:
                                if v != np.nan:
                                    velocities_max.append(v)
                                    velocities_min.append(v)
                                else:
                                    velocities_max.append(-float("inf"))
                                    velocities_min.append(float("inf"))
                            vmax = max(velocities_max)
                            vmin = min(velocities_min)
                            if vmax != -float("inf") and vmin != float("inf"):
                                if np.argmax(velocities_max) != np.argmin(
                                    velocities_min
                                ):
                                    t = (
                                        abs(
                                            np.argmax(velocities_max)
                                            - np.argmin(velocities_min)
                                        )
                                        / self.fps
                                    )
                                    self.all_behavior_parameters["acceleration"][i][
                                        n
                                    ] = (vmax - vmin) / t
                            self.all_behavior_parameters["distance"][
                                i
                            ] += distance_traveled
                            self.all_behavior_parameters["speed"][i][n] = speed

                        if "3 areal parameters" in parameter_to_analyze:
                            mask = np.zeros_like(self.background)
                            contour = self.animal_contours[i][n]
                            if contour is not None:
                                cv2.drawContours(
                                    mask, [contour], 0, (255, 255, 255), -1
                                )
                                mask = cv2.cvtColor(mask, cv2.COLOR_BGR2GRAY)
                                area_diffs = []
                                for ct in self.animal_contours[i][
                                    n - self.length + 1 : n + 1
                                ]:
                                    if ct is not None:
                                        prev_mask = np.zeros_like(self.background)
                                        cv2.drawContours(
                                            prev_mask, [ct], 0, (255, 255, 255), -1
                                        )
                                        prev_mask = cv2.cvtColor(
                                            prev_mask, cv2.COLOR_BGR2GRAY
                                        )
                                        xlf1, ybt1, w, h = cv2.boundingRect(contour)
                                        xrt1 = xlf1 + w
                                        ytp1 = ybt1 + h
                                        xlf2, ybt2, w, h = cv2.boundingRect(ct)
                                        xrt2 = xlf2 + w
                                        ytp2 = ybt2 + h
                                        xlf = min(xlf1, xlf2)
                                        xrt = max(xrt1, xrt2)
                                        ybt = min(ybt1, ybt2)
                                        ytp = max(ytp1, ytp2)
                                        curr_mask = mask[ybt:ytp, xlf:xrt]
                                        prev_mask = prev_mask[ybt:ytp, xlf:xrt]
                                        diff = cv2.bitwise_xor(prev_mask, curr_mask)
                                        # diff[diff!=0]=255
                                        area_diff = (np.sum(diff) / 255) / (
                                            np.sum(prev_mask) / 255
                                        )
                                        area_diffs.append(area_diff)
                                if len(area_diffs) > 0:
                                    magnitude_area = max(area_diffs)
                                    vigor_area = magnitude_area / (
                                        (self.length - np.argmax(area_diffs)) / self.fps
                                    )
                                    intensity_area = sum(area_diffs) / (
                                        self.length / self.fps
                                    )
                                    if magnitude_area > 0:
                                        self.all_behavior_parameters["magnitude_area"][
                                            i
                                        ][n] = magnitude_area
                                    if vigor_area > 0:
                                        self.all_behavior_parameters["vigor_area"][i][
                                            n
                                        ] = vigor_area
                                    if intensity_area > 0:
                                        self.all_behavior_parameters["intensity_area"][
                                            i
                                        ][n] = intensity_area

                    n += 1

                if self.categorize_behavior is True:
                    if "duration" in parameter_to_analyze:
                        for behavior_name in self.all_behavior_parameters:
                            if (
                                self.all_behavior_parameters[behavior_name]["duration"][
                                    i
                                ]
                                != 0.0
                            ):
                                self.all_behavior_parameters[behavior_name]["duration"][
                                    i
                                ] += round(self.length / self.fps, 2)
                            else:
                                self.all_behavior_parameters[behavior_name]["duration"][
                                    i
                                ] = "NA"

                    if "4 locomotion parameters" in parameter_to_analyze:
                        for behavior_name in self.all_behavior_parameters:
                            if (
                                self.all_behavior_parameters[behavior_name]["distance"][
                                    i
                                ]
                                == 0.0
                            ):
                                self.all_behavior_parameters[behavior_name]["distance"][
                                    i
                                ] = "NA"

    def export_results(self, normalize_distance=True, parameter_to_analyze=[]):
        print("Quantifying behaviors...")
        print(datetime.datetime.now())

        self.analyze_parameters(
            normalize_distance=normalize_distance,
            parameter_to_analyze=parameter_to_analyze,
        )

        print("Behavioral quantification completed!")

        print("Exporting results...")
        print(datetime.datetime.now())

        if self.categorize_behavior is True:
            events_df = pd.DataFrame(self.event_probability, index=self.all_time)
            events_df.to_excel(
                os.path.join(self.results_path, "all_event_probability.xlsx"),
                float_format="%.2f",
                index_label="time/ID",
            )

        all_parameters = []

        if self.categorize_behavior is True:
            all_parameters.append("probability")

        if "count" in parameter_to_analyze:
            all_parameters.append("count")
        if "duration" in parameter_to_analyze:
            all_parameters.append("duration")
        if "latency" in parameter_to_analyze:
            all_parameters.append("latency")
        if "3 length parameters" in parameter_to_analyze:
            all_parameters += ["intensity_length", "magnitude_length", "vigor_length"]
        if "3 areal parameters" in parameter_to_analyze:
            all_parameters += ["intensity_area", "magnitude_area", "vigor_area"]
        if "4 locomotion parameters" in parameter_to_analyze:
            all_parameters += ["acceleration", "distance", "speed", "velocity"]

        if self.categorize_behavior is True:
            for behavior_name in self.all_behavior_parameters:
                os.makedirs(
                    os.path.join(self.results_path, behavior_name), exist_ok=True
                )

                summary = []

                for parameter_name in all_parameters:
                    if parameter_name in ["count", "duration", "distance", "latency"]:
                        summary.append(
                            pd.DataFrame.from_dict(
                                self.all_behavior_parameters[behavior_name][
                                    parameter_name
                                ],
                                orient="index",
                                columns=[parameter_name],
                            ).reset_index(drop=True)
                        )
                    else:
                        individual_df = pd.DataFrame.from_dict(
                            self.all_behavior_parameters[behavior_name][parameter_name],
                            orient="index",
                            columns=self.all_time,
                        )
                        if parameter_name != "probability":
                            summary.append(
                                individual_df.mean(axis=1, skipna=True)
                                .to_frame()
                                .reset_index(drop=True)
                                .rename(columns={0: parameter_name + "_mean"})
                            )
                            summary.append(
                                individual_df.max(axis=1, skipna=True)
                                .to_frame()
                                .reset_index(drop=True)
                                .rename(columns={0: parameter_name + "_max"})
                            )
                            summary.append(
                                individual_df.min(axis=1, skipna=True)
                                .to_frame()
                                .reset_index(drop=True)
                                .rename(columns={0: parameter_name + "_min"})
                            )
                        individual_df = pd.DataFrame(
                            self.all_behavior_parameters[behavior_name][parameter_name],
                            index=self.all_time,
                        )
                        individual_df.to_excel(
                            os.path.join(
                                self.results_path,
                                behavior_name,
                                parameter_name + ".xlsx",
                            ),
                            float_format="%.2f",
                            index_label="time/ID",
                        )

                if len(summary) >= 1:
                    pd.concat(summary, axis=1).to_excel(
                        os.path.join(
                            self.results_path, behavior_name, "all_summary.xlsx"
                        ),
                        float_format="%.2f",
                        index_label="ID/parameter",
                    )

        else:
            summary = []

            for parameter_name in all_parameters:
                if parameter_name == "distance":
                    summary.append(
                        pd.DataFrame.from_dict(
                            self.all_behavior_parameters[parameter_name],
                            orient="index",
                            columns=["distance"],
                        ).reset_index(drop=True)
                    )
                else:
                    individual_df = pd.DataFrame.from_dict(
                        self.all_behavior_parameters[parameter_name],
                        orient="index",
                        columns=self.all_time,
                    )
                    summary.append(
                        individual_df.mean(axis=1, skipna=True)
                        .to_frame()
                        .reset_index(drop=True)
                        .rename(columns={0: parameter_name + "_mean"})
                    )
                    summary.append(
                        individual_df.max(axis=1, skipna=True)
                        .to_frame()
                        .reset_index(drop=True)
                        .rename(columns={0: parameter_name + "_max"})
                    )
                    summary.append(
                        individual_df.min(axis=1, skipna=True)
                        .to_frame()
                        .reset_index(drop=True)
                        .rename(columns={0: parameter_name + "_min"})
                    )
                    individual_df = pd.DataFrame(
                        self.all_behavior_parameters[parameter_name],
                        index=self.all_time,
                    )
                    individual_df.to_excel(
                        os.path.join(self.results_path, parameter_name + ".xlsx"),
                        float_format="%.2f",
                        index_label="time/ID",
                    )

            if len(summary) >= 1:
                pd.concat(summary, axis=1).to_excel(
                    os.path.join(self.results_path, "all_summary.xlsx"),
                    float_format="%.2f",
                    index_label="ID/parameter",
                )

        print("All results exported in: " + str(self.results_path))

    def generate_data(self, background_free=True, skip_redundant=1):
        print("Generating behavior examples...")
        print(datetime.datetime.now())

        background = self.background
        background_low = self.background_low
        background_high = self.background_high
        if self.animal_vs_bg == 1:
            background = np.uint8(255 - background)
            background_low = np.uint8(255 - background_low)
            background_high = np.uint8(255 - background_high)

        capture = cv2.VideoCapture(self.path_to_video)
        frame_count = frame_count_analyze = 0
        temp_frames = deque(maxlen=self.length)

        for i in range(self.animal_number):
            os.makedirs(
                os.path.join(self.results_path, "examples", str(i)), exist_ok=True
            )

        if self.t is not None and self.length is not None:
            start_t = round((self.t - self.length / self.fps), 2)
        else:
            start_t = 0.0

        if self.duration == 0:
            end_t = float("inf")
        else:
            end_t = start_t + self.duration

        while True:
            retval, frame = capture.read()
            time = round((frame_count + 1) / self.fps, 2)

            if time >= end_t or frame is None:
                break

            if time >= start_t:
                if self.framewidth is not None:
                    frame = cv2.resize(
                        frame,
                        (self.framewidth, self.frameheight),
                        interpolation=cv2.INTER_AREA,
                    )

                temp_frames.append(frame)

                (contours, centers, heights, inners, blobs) = contour_frame(
                    frame,
                    self.animal_number,
                    background,
                    background_low,
                    background_high,
                    self.delta,
                    self.animal_area,
                    animal_vs_bg=self.animal_vs_bg,
                    include_bodyparts=self.include_bodyparts,
                    animation_analyzer=False,
                    kernel=self.kernel,
                )

                if len(contours) > 0:
                    self.track_animal(
                        frame_count_analyze,
                        contours,
                        centers,
                        heights,
                        inners=inners,
                        blobs=None,
                    )

                    if (
                        frame_count_analyze >= self.length
                        and frame_count_analyze % skip_redundant == 0
                    ):
                        for n in self.animal_centers:
                            h = w = 0
                            animation = []
                            for i, f in enumerate(temp_frames):
                                contour = self.animal_contours[n][
                                    frame_count_analyze
                                    - self.length
                                    + 1 : frame_count_analyze + 1
                                ][i]
                                if contour is None:
                                    blob = np.zeros_like(f)
                                else:
                                    blob = extract_blob_background(
                                        f,
                                        self.animal_contours[n][
                                            frame_count_analyze
                                            - self.length
                                            + 1 : frame_count_analyze + 1
                                        ],
                                        contour=contour,
                                        channel=3,
                                        background_free=background_free,
                                    )
                                    h, w = blob.shape[:2]
                                animation.append(blob)

                            if h > 0:
                                if self.include_bodyparts is True:
                                    animation_name = (
                                        os.path.splitext(self.basename)[0]
                                        + "_"
                                        + str(n)
                                        + "_"
                                        + str(frame_count_analyze)
                                        + "_len"
                                        + str(self.length)
                                        + "_std"
                                        + str(self.std)
                                        + ".avi"
                                    )
                                    pattern_image_name = (
                                        os.path.splitext(self.basename)[0]
                                        + "_"
                                        + str(n)
                                        + "_"
                                        + str(frame_count_analyze)
                                        + "_len"
                                        + str(self.length)
                                        + "_std"
                                        + str(self.std)
                                        + ".jpg"
                                    )
                                    pattern_image = generate_patternimage(
                                        self.background,
                                        self.animal_contours[n][
                                            frame_count_analyze
                                            - self.length
                                            + 1 : frame_count_analyze + 1
                                        ],
                                        inners=self.animal_inners[n],
                                        std=self.std,
                                    )
                                else:
                                    animation_name = (
                                        os.path.splitext(self.basename)[0]
                                        + "_"
                                        + str(n)
                                        + "_"
                                        + str(frame_count_analyze)
                                        + "_len"
                                        + str(self.length)
                                        + ".avi"
                                    )
                                    pattern_image_name = (
                                        os.path.splitext(self.basename)[0]
                                        + "_"
                                        + str(n)
                                        + "_"
                                        + str(frame_count_analyze)
                                        + "_len"
                                        + str(self.length)
                                        + ".jpg"
                                    )
                                    pattern_image = generate_patternimage(
                                        self.background,
                                        self.animal_contours[n][
                                            frame_count_analyze
                                            - self.length
                                            + 1 : frame_count_analyze + 1
                                        ],
                                        inners=None,
                                        std=0,
                                    )

                                path_animation = os.path.join(
                                    self.results_path,
                                    "examples",
                                    str(n),
                                    animation_name,
                                )
                                path_pattern_image = os.path.join(
                                    self.results_path,
                                    "examples",
                                    str(n),
                                    pattern_image_name,
                                )

                                writer = cv2.VideoWriter(
                                    path_animation,
                                    cv2.VideoWriter_fourcc(*"MJPG"),
                                    self.fps / 5,
                                    (w, h),
                                    True,
                                )
                                for blob in animation:
                                    writer.write(
                                        cv2.resize(
                                            blob, (w, h), interpolation=cv2.INTER_AREA
                                        )
                                    )
                                writer.release()

                                cv2.imwrite(path_pattern_image, pattern_image)

                frame_count_analyze += 1

            frame_count += 1

        capture.release()

        print("Behavior example generation completed!")

    def generate_data_interact_basic(self, background_free=True, skip_redundant=1):
        print("Generating behavior examples...")
        print(datetime.datetime.now())

        background = self.background
        background_low = self.background_low
        background_high = self.background_high
        if self.animal_vs_bg == 1:
            background = np.uint8(255 - background)
            background_low = np.uint8(255 - background_low)
            background_high = np.uint8(255 - background_high)

        capture = cv2.VideoCapture(self.path_to_video)
        frame_count = frame_count_analyze = 0
        temp_frames = deque(maxlen=self.length)
        temp_contours = deque(maxlen=self.length)
        temp_inners = deque(maxlen=self.length)
        animation = deque(maxlen=self.length)
        os.makedirs(os.path.join(self.results_path, "examples", "0"), exist_ok=True)

        start_t = round((self.t - self.length / self.fps), 2)
        if start_t < 0:
            start_t = 0.00
        if self.duration == 0:
            end_t = float("inf")
        else:
            end_t = start_t + self.duration

        while True:
            retval, frame = capture.read()
            time = round((frame_count + 1) / self.fps, 2)

            if time >= end_t or frame is None:
                break

            if time >= start_t:
                if self.framewidth is not None:
                    frame = cv2.resize(
                        frame,
                        (self.framewidth, self.frameheight),
                        interpolation=cv2.INTER_AREA,
                    )

                temp_frames.append(frame)

                (contours, centers, heights, inners, blobs) = contour_frame(
                    frame,
                    self.animal_number,
                    background,
                    background_low,
                    background_high,
                    self.delta,
                    self.animal_area,
                    animal_vs_bg=self.animal_vs_bg,
                    include_bodyparts=self.include_bodyparts,
                    animation_analyzer=False,
                    kernel=self.kernel,
                )

                if len(contours) == 0:
                    temp_contours.append(None)
                    temp_inners.append(None)
                    animation.append(np.zeros_like(self.background))

                else:
                    temp_contours.append(contours)
                    temp_inners.append(inners)

                    if (
                        frame_count_analyze >= self.length
                        and frame_count_analyze % skip_redundant == 0
                    ):
                        (y_bt, y_tp, x_lf, x_rt) = crop_frame(
                            frame, functools.reduce(operator.iconcat, temp_contours, [])
                        )

                        h = w = 0
                        for i, f in enumerate(temp_frames):
                            if temp_contours[i] is None:
                                blob = np.zeros_like(f)
                            else:
                                blob = extract_blob_all(
                                    f,
                                    y_bt,
                                    y_tp,
                                    x_lf,
                                    x_rt,
                                    contours=temp_contours[i],
                                    channel=3,
                                    background_free=background_free,
                                )
                                h, w = blob.shape[:2]
                            animation.append(blob)

                        if h > 0:
                            if self.include_bodyparts is True:
                                animation_name = (
                                    os.path.splitext(self.basename)[0]
                                    + "_"
                                    + str(frame_count_analyze)
                                    + "_len"
                                    + str(self.length)
                                    + "_std"
                                    + str(self.std)
                                    + "_itbs.avi"
                                )
                                pattern_image_name = (
                                    os.path.splitext(self.basename)[0]
                                    + "_"
                                    + str(frame_count_analyze)
                                    + "_len"
                                    + str(self.length)
                                    + "_std"
                                    + str(self.std)
                                    + "_itbs.jpg"
                                )
                                pattern_image = generate_patternimage_all(
                                    frame,
                                    y_bt,
                                    y_tp,
                                    x_lf,
                                    x_rt,
                                    temp_contours,
                                    temp_inners,
                                    std=self.std,
                                )
                            else:
                                animation_name = (
                                    os.path.splitext(self.basename)[0]
                                    + "_"
                                    + str(frame_count_analyze)
                                    + "_len"
                                    + str(self.length)
                                    + "_itbs.avi"
                                )
                                pattern_image_name = (
                                    os.path.splitext(self.basename)[0]
                                    + "_"
                                    + str(frame_count_analyze)
                                    + "_len"
                                    + str(self.length)
                                    + "_itbs.jpg"
                                )
                                pattern_image = generate_patternimage_all(
                                    frame,
                                    y_bt,
                                    y_tp,
                                    x_lf,
                                    x_rt,
                                    temp_contours,
                                    temp_inners,
                                    std=0,
                                )

                            path_animation = os.path.join(
                                self.results_path, "examples", "0", animation_name
                            )
                            path_pattern_image = os.path.join(
                                self.results_path, "examples", "0", pattern_image_name
                            )

                            writer = cv2.VideoWriter(
                                path_animation,
                                cv2.VideoWriter_fourcc(*"MJPG"),
                                self.fps / 5,
                                (w, h),
                                True,
                            )
                            for blob in animation:
                                writer.write(
                                    cv2.resize(
                                        blob, (w, h), interpolation=cv2.INTER_AREA
                                    )
                                )
                            writer.release()

                            cv2.imwrite(path_pattern_image, pattern_image)

                frame_count_analyze += 1

            frame_count += 1

        capture.release()

        print("Behavior example generation completed!")



"""
Copyright (C)
This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program. If not, see https://tldrlegal.com/license/gnu-general-public-license-v3-(gpl-3)#fulltext. 
For license issues, please contact:
Dr. Bing Ye
Life Sciences Institute
University of Michigan
210 Washtenaw Avenue, Room 5403
Ann Arbor, MI 48109-2216
USA
Email: bingye@umich.edu
"""


import os

import pandas as pd
import scikit_posthocs as sp
from scipy import stats


class data_mining:
    def __init__(
        self,
        data_in,
        control_in=None,
        paired_in=False,
        result_path_in=None,
        pval_in=0.05,
        file_names_in=None,
    ):
        self.data = data_in
        self.control = control_in
        self.paired = paired_in
        self.pval = pval_in
        self.writer = pd.ExcelWriter(
            os.path.join(result_path_in, "data_mining_results.xlsx")
        )
        self.file_names = file_names_in

    def normal(self, dataset):
        normal = True
        for i in dataset:
            if len(dataset) >= 3:
                if stats.shapiro(i).pvalue < self.pval:
                    normal = False
        return normal

    def two_groups(self):
        if self.control != None:
            self.data.insert(0, self.control)
        for behavior in self.data[0]:
            startrow = 0
            print(behavior)
            for parameter in self.data[0][behavior]:
                parameters = []
                pvalues = []
                setA = self.data[0][behavior][parameter].dropna()
                setB = self.data[1][behavior][parameter].dropna()
                dataset = [setA, setB]
                if self.normal(dataset):
                    if self.paired:
                        test = "paired t-test"
                        result = stats.ttest_rel(setA, setB)
                    else:
                        test = "unpaired t-test"
                        result = stats.ttest_ind(setA, setB)
                else:
                    if self.paired:
                        test = "Wilcoxon test"
                        result = stats.wilcoxon(setA, setB)
                    else:
                        test = "Mann Whitney U test"
                        result = stats.mannwhitneyu(setA, setB)
                if result.pvalue <= self.pval:
                    print("\t", parameter)
                    print("\t", "performing", test)
                    print("\t" * 2, "p-value: ", result.pvalue)
                    parameters.append(parameter)
                    pvalues.append(result.pvalue)

                    stat_info = pd.DataFrame(
                        {"p-value": pvalues}, index=parameters
                    )  # parameter and associated p-value
                    stat_info.to_excel(
                        self.writer, sheet_name=behavior, startrow=startrow
                    )

                    significant_data = pd.DataFrame(
                        (setA, setB)
                    ).transpose()  # print out the datasets with significant findings
                    significant_data.columns = self.file_names
                    significant_data.to_excel(
                        self.writer, sheet_name=behavior, startrow=startrow, startcol=3
                    )

                    sheet = self.writer.sheets[behavior]
                    sheet.write_string(startrow, 0, test)
                    startrow += significant_data.shape[0] + 2

    def multiple_groups(self):
        for behavior in self.data[0]:
            startrow = 0
            print(behavior)
            for parameter in self.data[0][behavior]:
                dataset = []
                parameters = []
                pvalues = []
                for i in range(len(self.data)):
                    currentSet = self.data[i][behavior][parameter].dropna()
                    dataset.append(currentSet)
                if self.control != None:
                    currentSet = self.control[behavior][parameter].dropna()
                    dataset.insert(0, currentSet)
                if self.normal(dataset):
                    test = "ANOVA"
                    result = stats.f_oneway(*dataset)
                else:
                    if self.paired:
                        test = "Friedman"
                        result = stats.friedmanchisquare(*dataset)
                    else:
                        test = "Kruskal Wallis"
                        result = stats.kruskal(*dataset)
                if result.pvalue <= self.pval:
                    pvalues = []
                    parameters = []
                    posthoc = None
                    posthoc_name = None
                    print("\t", parameter)
                    print("\t", "performing", test)
                    print("\t" * 2, "p-value: ", result.pvalue)

                    pvalues.append(result.pvalue)
                    parameters.append(parameter)

                    stat_info = pd.DataFrame(
                        {"p-value": pvalues}, index=parameters
                    )  # parameter and associated p-value
                    stat_info.to_excel(
                        self.writer, sheet_name=behavior, startrow=startrow
                    )

                    significant_data = pd.DataFrame(
                        dataset
                    ).transpose()  # print out the datasets with significant findings
                    significant_data.columns = self.file_names
                    significant_data.to_excel(
                        self.writer, sheet_name=behavior, startrow=startrow, startcol=8
                    )

                    if test == "ANOVA":
                        if self.control == None:
                            tukey = stats.tukey_hsd(*dataset)
                            print(tukey)

                            posthoc_name = "Tukey"
                            posthoc = pd.DataFrame(tukey)
                        else:
                            dataset_exp = dataset[1:]
                            dunnett = stats.dunnett(*dataset_exp, control=self.control)
                            dunnett.columns = self.file_names[1:]
                            dunnett.index = self.file_names[1:]
                            print("\t" * 2, "Dunnett's post-hoc results:")
                            print(
                                "\t" * 2 + dunnett.to_string().replace("\n", "\n\t\t")
                            )

                            posthoc_name = "Dunnett"
                            posthoc = pd.DataFrame(dunnett)
                    else:
                        dunn = sp.posthoc_dunn(dataset)
                        dunn.columns = self.file_names
                        dunn.index = self.file_names
                        print("\t" * 2, "Dunn's post-hoc results:")
                        print("\t" * 2 + dunn.to_string().replace("\n", "\n\t\t"))

                        posthoc_name = "Dunn"
                        posthoc = pd.DataFrame(dunn)

                    posthoc.to_excel(
                        self.writer, sheet_name=behavior, startrow=startrow, startcol=3
                    )
                    sheet = self.writer.sheets[behavior]
                    sheet.write_string(startrow, 0, test)
                    sheet.write_string(startrow, 3, posthoc_name + "'s post-hoc")
                    startrow += significant_data.shape[0] + 2

    def statistical_analysis(self):
        if (len(self.data) == 2 and self.control == None) or (
            len(self.data) == 1 and self.control != None
        ):  # tests for two groups only
            self.two_groups()
        else:  # tests for 3+ groups
            self.multiple_groups()
        self.writer.close()
        print("Data mining for statistical analysis completed!")
